#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import random
import string
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Optional, Union

from boto3 import Session

import seedfarmer.errors
import seedfarmer.services._cloudwatch as cloudwatch
import seedfarmer.services._codebuild as codebuild
import seedfarmer.services._s3 as s3
from seedfarmer.error_handler import log_error_safely

_logger: logging.Logger = logging.getLogger(__name__)


def _print_codebuild_logs(
    events: List[cloudwatch.CloudWatchEvent],
    codebuild_log_callback: Optional[Callable[[str], None]] = None,
) -> None:
    for event in events:
        msg = event.message[:-1] if event.message.endswith("\n") else event.message
        _logger.debug("[CODEBUILD] %s", msg)
        if codebuild_log_callback:
            codebuild_log_callback(msg)


def _wait_execution(
    build_id: str,
    stream_name_prefix: str,
    codebuild_log_callback: Optional[Callable[[str], None]] = None,
    session: Optional[Union[Callable[[], Session], Session]] = None,
) -> Optional[codebuild.BuildInfo]:
    start_time: Optional[datetime] = None
    stream_name: Optional[str] = None
    status: Optional[codebuild.BuildInfo] = None
    for status in codebuild.wait(build_id=build_id, session=session):
        if status.logs.enabled and status.logs.group_name:
            if stream_name is None:
                stream_name = cloudwatch.get_stream_name_by_prefix(
                    group_name=status.logs.group_name, prefix=f"{stream_name_prefix}/", session=session
                )
            if stream_name is not None:
                events = cloudwatch.get_log_events(
                    group_name=status.logs.group_name, stream_name=stream_name, start_time=start_time, session=session
                )
                _print_codebuild_logs(events=events.events, codebuild_log_callback=codebuild_log_callback)
                if events.last_timestamp is not None:
                    start_time = events.last_timestamp + timedelta(milliseconds=1)
    return status


def _execute_codebuild(
    stack_outputs: Dict[str, str],
    bundle_location: str,
    execution_id: str,
    buildspec: Dict[str, Any],
    timeout: int,
    overrides: Optional[Dict[str, Any]] = None,
    codebuild_log_callback: Optional[Callable[[str], None]] = None,
    session: Optional[Union[Callable[[], Session], Session]] = None,
    yaml_dumper: Optional[Any] = None,  # Accepts ruamel.yaml.YAML instance or PyYAML dump function
) -> Optional[codebuild.BuildInfo]:
    _logger.debug("bundle_location: %s", bundle_location)
    stream_name_prefix = f"codeseeder-{execution_id}"  # (LEGACY)
    _logger.debug("stream_name_prefix: %s", stream_name_prefix)
    build_id = codebuild.start(
        project_name=stack_outputs["CodeBuildProject"],
        stream_name=stream_name_prefix,
        bundle_location=bundle_location,
        buildspec=buildspec,
        timeout=timeout,
        overrides=overrides,
        session=session,
        yaml_dumper=yaml_dumper,
    )
    return _wait_execution(
        build_id=build_id,
        stream_name_prefix=stream_name_prefix,
        codebuild_log_callback=codebuild_log_callback,
        session=session,
    )


def run(
    stack_outputs: Dict[str, str],
    bundle_path: str,
    buildspec: Dict[str, Any],
    timeout: int,
    overrides: Optional[Dict[str, Any]] = None,
    codebuild_log_callback: Optional[Callable[[str], None]] = None,
    session: Optional[Union[Callable[[], Session], Session]] = None,
    bundle_id: Optional[str] = None,
    prebuilt_bundle: Optional[str] = None,
    yaml_dumper: Optional[Any] = None,  # Accepts ruamel.yaml.YAML instance or PyYAML dump function
) -> Optional[codebuild.BuildInfo]:
    execution_id = "".join(random.choice(string.ascii_lowercase) for i in range(8))

    if prebuilt_bundle:
        prebuilt_bundle = prebuilt_bundle[5:] if prebuilt_bundle.startswith("s3://") else prebuilt_bundle
        o = prebuilt_bundle.split("/", 1)
        loc = f"{o[0]}/{o[1]}"
    else:
        try:
            key: str = (
                f"seedfarmer/{bundle_id}/{execution_id}/bundle.zip"
                if bundle_id
                else f"seedfarmer/{execution_id}/bundle.zip"
            )
            bucket = stack_outputs["Bucket"]
            s3.delete_objects(bucket=bucket, keys=[key], session=session)
            s3.upload_file(src=bundle_path, bucket=bucket, key=key, session=session)
            loc = f"{bucket}/{key}"
        except Exception as e:
            log_error_safely(
                _logger, e, "Failed to upload deployment bundle to S3. Verify SeedKit has deployed successfully."
            )
            # Show the actual error to user without stacktrace
            _logger.error(f"S3 upload failed: {e}")
            raise seedfarmer.errors.RemoteDeploymentRuntimeError(f"S3 upload failed: {e}")

    try:
        build_info = _execute_codebuild(
            stack_outputs=stack_outputs,
            bundle_location=loc,
            execution_id=execution_id,
            buildspec=buildspec,
            codebuild_log_callback=codebuild_log_callback,
            timeout=timeout,
            overrides=overrides,
            session=session,
            yaml_dumper=yaml_dumper,
        )
    except Exception as e:
        log_error_safely(_logger, e, "CodeBuild execution failed")
        # Show the actual error to user without stacktrace
        _logger.error(f"CodeBuild execution failed: {e}")
        raise seedfarmer.errors.RemoteDeploymentRuntimeError(f"CodeBuild execution failed: {e}")
    finally:
        # Clean up S3 bundle even if execution failed (unless it's prebuilt)
        if not prebuilt_bundle:
            try:
                s3.delete_objects(bucket=bucket, keys=[key], session=session)
            except Exception as e:
                log_error_safely(_logger, e, f"Failed to clean up S3 bundle: {bucket}/{key}")
                # Don't raise here - this is just cleanup

    return build_info



from typing import Dict, Optional


class InvalidPathError(RuntimeError):
    def __init__(self, *args: str):
        super().__init__(*args)


class InvalidManifestError(RuntimeError):
    def __init__(self, *args: str, error_file_path: Optional[str] = None):
        super().__init__(*args)


class InvalidConfigurationError(RuntimeError):
    def __init__(self, *args: str):
        super().__init__(*args)


class InvalidSessionError(RuntimeError):
    def __init__(self, *args: str, error_message: Optional[str] = None):
        super().__init__(*args)


class ModuleDeploymentError(RuntimeError):
    def __init__(self, *args: str, error_message: Optional[str] = None):
        super().__init__(*args)


class SeedFarmerException(Exception):
    def __init__(self, *args: str, error_message: Optional[str] = None):
        super().__init__(*args)


class RemoteDeploymentRuntimeError(RuntimeError):
    def __init__(self, *args: str, error_info: Optional[Dict[str, str]] = None):
        super().__init__(*args)
        self.error_info = error_info



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import sys
from functools import wraps
from typing import Any, Callable, NoReturn, Optional, TypeVar

import botocore.exceptions
import click

from seedfarmer.errors.seedfarmer_errors import (
    InvalidConfigurationError,
    InvalidManifestError,
    InvalidPathError,
    InvalidSessionError,
    ModuleDeploymentError,
    RemoteDeploymentRuntimeError,
    SeedFarmerException,
)

_logger: logging.Logger = logging.getLogger(__name__)

F = TypeVar("F", bound=Callable[..., Any])


def print_error_message(message: str, color: str = "red") -> None:
    """
    Simple error message printer to avoid circular imports.
    """
    if color == "red":
        click.echo(click.style(message, fg="red", bold=True))
    elif color == "yellow":
        click.echo(click.style(message, fg="yellow", bold=True))
    else:
        click.echo(click.style(message, bold=True))


def handle_seedfarmer_error(error: Exception, context: Optional[str] = None) -> NoReturn:
    """
    Handle SeedFarmer specific errors with user-friendly messages.

    Args:
        error: The exception that occurred
        context: Optional context about when the error occurred
    """
    error_prefix = f"[{context}] " if context else ""

    if isinstance(error, InvalidPathError):
        print_error_message(f"{error_prefix}Path Error: {error}", "red")
        _logger.debug("Full error details: %s", error, exc_info=True)

    elif isinstance(error, InvalidManifestError):
        print_error_message(f"{error_prefix}Manifest Error: {error}", "red")
        print_error_message("Please check your deployment manifest syntax and structure.", "yellow")
        _logger.debug("Full error details: %s", error, exc_info=True)

    elif isinstance(error, InvalidConfigurationError):
        print_error_message(f"{error_prefix}Configuration Error: {error}", "red")
        print_error_message("Please verify your configuration settings.", "yellow")
        _logger.debug("Full error details: %s", error, exc_info=True)

    elif isinstance(error, ModuleDeploymentError):
        print_error_message(f"{error_prefix}Deployment Error: {error}", "red")
        print_error_message("Check the deployment logs for more details.", "yellow")
        _logger.debug("Full error details: %s", error, exc_info=True)

    elif isinstance(error, InvalidSessionError):
        print_error_message(f"{error_prefix}Session Error: {error}", "red")
        print_error_message("Please check your AWS credentials and permissions.", "yellow")
        _logger.debug("Full error details: %s", error, exc_info=True)

    elif isinstance(error, RemoteDeploymentRuntimeError):
        print_error_message(f"{error_prefix}Remote Deployment Error: {error}", "red")
        _logger.debug("Full error details: %s", error, exc_info=True)

    else:
        # Handle the error as a generic exception
        handle_generic_error(error, context)
        return  # handle_generic_error already calls sys.exit

    sys.exit(1)


def handle_generic_error(error: Exception, context: Optional[str] = None) -> NoReturn:
    """
    Handle generic errors with user-friendly messages.

    Args:
        error: The exception that occurred
        context: Optional context about when the error occurred
    """
    error_prefix = f"[{context}] " if context else ""

    if isinstance(error, botocore.exceptions.NoCredentialsError):
        print_error_message(f"{error_prefix}AWS Credentials Error: No AWS credentials found", "red")
        print_error_message(
            "Please configure your AWS credentials using 'aws configure' or environment variables.", "yellow"
        )
        _logger.debug("Full error details: %s", error, exc_info=True)

    elif isinstance(error, botocore.exceptions.ClientError):
        error_code = error.response.get("Error", {}).get("Code", "Unknown")
        error_message = error.response.get("Error", {}).get("Message", str(error))
        print_error_message(f"{error_prefix}AWS Error ({error_code}): {error_message}", "red")

        # Provide specific guidance for common AWS errors
        if error_code == "AccessDenied":
            print_error_message("Please check your AWS permissions and IAM roles.", "yellow")
        elif error_code == "UnauthorizedOperation":
            print_error_message("Please check your AWS permissions for this operation.", "yellow")
        elif error_code in ["NoSuchBucket", "BucketNotFound"]:
            print_error_message("The specified S3 bucket was not found.", "yellow")
        elif error_code == "InvalidParameterValue":
            print_error_message("Please check the parameter values in your configuration.", "yellow")

        _logger.debug("Full error details: %s", error, exc_info=True)

    elif isinstance(error, FileNotFoundError):
        print_error_message(f"{error_prefix}File Not Found: {error.filename or 'Unknown file'}", "red")
        print_error_message("Please check that all required files exist and paths are correct.", "yellow")
        _logger.debug("Full error details: %s", error, exc_info=True)

    elif isinstance(error, PermissionError):
        print_error_message(f"{error_prefix}Permission Error: {error}", "red")
        print_error_message("Please check file and directory permissions.", "yellow")
        _logger.debug("Full error details: %s", error, exc_info=True)

    elif isinstance(error, KeyboardInterrupt):
        print_error_message("\nOperation cancelled by user.", "yellow")
        _logger.debug("Operation interrupted by user")

    else:
        # For truly unexpected errors, show a generic message but log the full details
        print_error_message(f"{error_prefix}Unexpected Error: {type(error).__name__}: {error}", "red")
        print_error_message("An unexpected error occurred. Please check the logs for more details.", "yellow")
        _logger.error("Unexpected error occurred", exc_info=True)

    sys.exit(1)


def safe_execute(context: str) -> Callable[[F], F]:
    """
    Decorator to safely execute functions with proper error handling.

    Args:
        context: Description of what operation is being performed

    Returns:
        Decorated function that handles errors gracefully
    """

    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return func(*args, **kwargs)
            except (
                InvalidPathError,
                InvalidManifestError,
                InvalidConfigurationError,
                ModuleDeploymentError,
                InvalidSessionError,
                RemoteDeploymentRuntimeError,
                SeedFarmerException,
            ) as e:
                handle_seedfarmer_error(e, context)
            except Exception as e:
                handle_generic_error(e, context)

        return wrapper  # type: ignore

    return decorator


def log_error_safely(logger: logging.Logger, error: Exception, message: str) -> None:
    """
    Log an error message without exposing sensitive stacktrace information to console output.

    Args:
        logger: Logger instance to use
        error: The exception that occurred
        message: User-friendly error message
    """
    # Log user-friendly message at error level (will show in console)
    logger.error(message)

    # Log full exception details only at debug level (hidden unless debug mode is on)
    logger.debug("Full error details: %s", error, exc_info=True)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import io
import logging
import time
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Callable, Dict, Iterable, List, NamedTuple, Optional, Union, cast

import botocore.exceptions
import yaml
from boto3 import Session

from seedfarmer.errors import InvalidConfigurationError
from seedfarmer.services._service_utils import boto3_client, try_it
from seedfarmer.utils import LiteralStr

_logger: logging.Logger = logging.getLogger(__name__)
_BUILD_WAIT_POLLING_DELAY: float = 5  # SECONDS


class BuildStatus(Enum):
    failed = "FAILED"
    fault = "FAULT"
    in_progress = "IN_PROGRESS"
    stopped = "STOPPED"
    succeeded = "SUCCEEDED"
    timed_out = "TIMED_OUT"


class BuildPhaseType(Enum):
    build = "BUILD"
    completed = "COMPLETED"
    download_source = "DOWNLOAD_SOURCE"
    finalizing = "FINALIZING"
    install = "INSTALL"
    post_build = "POST_BUILD"
    pre_build = "PRE_BUILD"
    provisioning = "PROVISIONING"
    queued = "QUEUED"
    submitted = "SUBMITTED"
    upload_artifacts = "UPLOAD_ARTIFACTS"


class BuildPhaseStatus(Enum):
    failed = "FAILED"
    fault = "FAULT"
    queued = "QUEUED"
    in_progress = "IN_PROGRESS"
    stopped = "STOPPED"
    succeeded = "SUCCEEDED"
    timed_out = "TIMED_OUT"


class BuildPhaseContext(NamedTuple):
    status_code: Optional[str]
    message: Optional[str]


class BuildPhase(NamedTuple):
    phase_type: BuildPhaseType
    status: Optional[BuildPhaseStatus]
    start_time: datetime
    end_time: Optional[datetime]
    duration_in_seconds: float
    contexts: List[BuildPhaseContext]


class BuildCloudWatchLogs(NamedTuple):
    enabled: bool
    group_name: Optional[str]
    stream_name: Optional[str]


class BuildInfo(NamedTuple):
    build_id: str
    status: BuildStatus
    current_phase: BuildPhaseType
    start_time: datetime
    end_time: Optional[datetime]
    duration_in_seconds: float
    phases: List[BuildPhase]
    logs: BuildCloudWatchLogs
    exported_env_vars: Dict[str, str]


def get_build_data(
    build_ids: List[str], session: Optional[Union[Callable[[], Session], Session]] = None
) -> Optional[Dict[str, Any]]:
    client = boto3_client(service_name="codebuild", session=session)

    try:
        return cast(Dict[str, Any], client.batch_get_builds(ids=build_ids))
    except Exception as e:
        _logger.error("An error occurred fetching the build info for %s - %s", build_ids, e)
        return None


def start(
    project_name: str,
    stream_name: str,
    bundle_location: str,
    buildspec: Dict[str, Any],
    timeout: int,
    overrides: Optional[Dict[str, Any]] = None,
    session: Optional[Union[Callable[[], Session], Session]] = None,
    yaml_dumper: Optional[Any] = None,  # Accepts ruamel.yaml.YAML instance or PyYAML dump function
) -> str:
    """Start a CodeBuild Project execution

    Parameters
    ----------
    project_name : str
        Name of the CodeBuild Project
    stream_name : str
        Name of the CloudWatch Logs Stream to read CodeBuild logs from
    bundle_location : str
        S3 Location of the zip bundle to use as Source to the CodeBuild execution
    buildspec : Dict[str, Any]
        BuildSpec to use for the CodeBuild execution
    timeout : int
        Timeout of the CodeBuild execution
    overrides : Optional[Dict[str, Any]], optional
        Additional overrides to set on the CodeBuild execution, by default None
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None

    Returns
    -------
    str
        The CodeBuild Build/Exectuion Id
    """
    # If no dumper is provided, use PyYAML's safe_dump
    if yaml_dumper is None:

        def default_pyyaml_dumper(data):  # type: ignore[no-untyped-def]
            return yaml.safe_dump(data, sort_keys=False, indent=4)

        yaml_dumper = default_pyyaml_dumper

    # ruamel.yaml.YAML instance: has a .dump method
    if hasattr(yaml_dumper, "dump"):
        stream = io.StringIO()
        yaml_dumper.dump(buildspec, stream)
        buildspec_yaml = stream.getvalue()
        stream.close()
    # PyYAML: function like yaml.dump or yaml.safe_dump
    elif callable(yaml_dumper):
        buildspec_yaml = yaml_dumper(buildspec)
    else:
        raise InvalidConfigurationError("yaml_dumper must be a ruamel.yaml.YAML instance or a PyYAML dump function.")

    client = boto3_client("codebuild", session=session)
    image_override: Optional[str] = overrides.get("imageOverride", None) if overrides else None
    image_pull_credentials: Optional[str] = None
    if image_override:
        if image_override.startswith("aws/"):
            image_pull_credentials = "CODEBUILD"
        else:
            image_pull_credentials = "SERVICE_ROLE"
            _logger.debug("Image Pull Credentials: %s", image_pull_credentials)

    _logger.debug("Overrides: %s", overrides)
    build_params = {
        "projectName": project_name,
        "sourceTypeOverride": "S3",
        "sourceLocationOverride": bundle_location,
        "buildspecOverride": buildspec_yaml,
        "timeoutInMinutesOverride": timeout,
        "privilegedModeOverride": True,
        "logsConfigOverride": {
            "cloudWatchLogs": {
                "status": "ENABLED",
                "groupName": f"/aws/codebuild/{project_name}",
                "streamName": stream_name,
            },
            "s3Logs": {"status": "DISABLED"},
        },
    }
    if image_pull_credentials:
        build_params["imagePullCredentialsTypeOverride"] = image_pull_credentials

    if overrides:
        build_params = {**build_params, **overrides}

    response = client.start_build(**build_params)  # type: ignore[arg-type]
    return response["build"]["id"]


def fetch_build_info(build_id: str, session: Optional[Union[Callable[[], Session], Session]] = None) -> BuildInfo:
    """Fetch info on a CodeBuild execution

    Parameters
    ----------
    build_id : str
        CodeBuild Execution/Build Id
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None

    Returns
    -------
    BuildInfo
        Info on the CodeBuild execution

    Raises
    ------
    RuntimeError
        If the Build Id is not found
    """
    client = boto3_client("codebuild", session=session)
    response: Dict[str, List[Dict[str, Any]]] = try_it(
        f=client.batch_get_builds, ex=botocore.exceptions.ClientError, ids=[build_id], max_num_tries=5
    )
    if not response["builds"]:
        raise RuntimeError(f"CodeBuild build {build_id} not found.")
    build = response["builds"][0]
    now = datetime.now(timezone.utc)
    log_enabled = True if build.get("logs", {}).get("cloudWatchLogs", {}).get("status") == "ENABLED" else False
    return BuildInfo(
        build_id=build_id,
        status=BuildStatus(value=build["buildStatus"]),
        current_phase=BuildPhaseType(value=build["currentPhase"]),
        start_time=build["startTime"],
        end_time=build.get("endTime", now),
        duration_in_seconds=(build.get("endTime", now) - build["startTime"]).total_seconds(),
        exported_env_vars={d["name"]: d["value"] for d in build.get("exportedEnvironmentVariables", [])},
        phases=[
            BuildPhase(
                phase_type=BuildPhaseType(value=p["phaseType"]),
                status=None if "phaseStatus" not in p else BuildPhaseStatus(value=p["phaseStatus"]),
                start_time=p["startTime"],
                end_time=p.get("endTime", now),
                duration_in_seconds=p.get("durationInSeconds"),
                contexts=[
                    BuildPhaseContext(status_code=p.get("statusCode"), message=p.get("message"))
                    for c in p.get("contexts", [])
                ],
            )
            for p in build["phases"]
        ],
        logs=BuildCloudWatchLogs(
            enabled=log_enabled,
            group_name=build["logs"]["cloudWatchLogs"].get("groupName") if log_enabled else None,
            stream_name=build["logs"]["cloudWatchLogs"].get("streamName") if log_enabled else None,
        ),
    )


def wait(build_id: str, session: Optional[Union[Callable[[], Session], Session]] = None) -> Iterable[BuildInfo]:
    """Wait for completion of a CodeBuild execution

    Parameters
    ----------
    build_id : str
        The CodeBuild Execution/Build Id
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None

    Returns
    -------
    Iterable[BuildInfo]
        Info on the CodeBuild execution

    Yields
    -------
    Iterator[Iterable[BuildInfo]]
        Info on the CodeBuild execution

    """
    build = fetch_build_info(build_id=build_id, session=session)
    while build.status is BuildStatus.in_progress:
        time.sleep(_BUILD_WAIT_POLLING_DELAY)

        last_phase = build.current_phase
        last_status = build.status
        build = fetch_build_info(build_id=build_id, session=session)

        if build.current_phase is not last_phase or build.status is not last_status:
            _logger.info("phase: %s %s (%s)", build.current_phase.value, build.build_id, build.status.value)

        yield build

    yield build


def generate_spec(
    cmds_install: Optional[List[Union[str, LiteralStr]]] = None,
    cmds_pre: Optional[List[Union[str, LiteralStr]]] = None,
    cmds_build: Optional[List[Union[str, LiteralStr]]] = None,
    cmds_post: Optional[List[Union[str, LiteralStr]]] = None,
    env_vars: Optional[Dict[str, str]] = None,
    exported_env_vars: Optional[List[str]] = None,
    runtime_versions: Optional[Dict[str, str]] = None,
    abort_phases_on_failure: bool = True,
) -> Dict[str, Any]:
    """Generate a BuildSpec for a CodeBuild execution

    Parameters
    ----------
    stack_outputs : Dict[str, str]
        The CloudFormation Stack Outputs from the Seedkit Stack where the CodeBuild Project was created
    cmds_install : Optional[List[str]], optional
        Additional commands to run during the Install phase of the CodeBuild execution, by default None
    cmds_pre : Optional[List[str]], optional
        Additional commands to run during the PreBuild phase of the CodeBuild execution, by default None
    cmds_build : Optional[List[str]], optional
        Additional commands to run during the Build phase of the CodeBuild execution, by default None
    cmds_post : Optional[List[str]], optional
        Additional commands to run during the PostBuild phase of the CodeBuild execution, by default None
    env_vars: Optional[Dict[str, str]], optional
        Environment variables to set in the CodeBuild execution, by default None
    exported_env_vars: Optional[List[str]], optional
        Environment variables to export from the CodeBuild execution, by default None
    pypi_mirror: Optional[str], optional
        Pypi mirror to use, by default None
    npm_mirror: Optional[str], optional
        NPM mirror to use, by default None

    Returns
    -------
    Dict[str, Any]
        A CodeBuild BuildSpec
    """
    # pre: List[str] = [] if cmds_pre is None else cmds_pre
    # build: List[str] = [] if cmds_build is None else cmds_build
    # post: List[str] = [] if cmds_post is None else cmds_post
    variables: Dict[str, str] = {} if env_vars is None else env_vars
    # exported_variables: List[str] = [] if exported_env_vars is None else exported_env_vars

    on_failure = "ABORT" if abort_phases_on_failure else "CONTINUE"
    return_spec: Dict[str, Any] = {
        "version": 0.2,
        "env": {"shell": "bash", "variables": variables, "exported-variables": exported_env_vars or []},
        "phases": {
            "install": {
                "commands": cmds_install or [],
                "on-failure": on_failure,
            },
            "pre_build": {
                "commands": cmds_pre or [],
                "on-failure": on_failure,
            },
            "build": {
                "commands": cmds_build or [],
                "on-failure": on_failure,
            },
            "post_build": {
                "commands": cmds_post or [],
                "on-failure": on_failure,
            },
        },
    }
    if runtime_versions:
        return_spec["phases"]["install"]["runtime-versions"] = runtime_versions

    _logger.debug(return_spec)
    return return_spec



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from typing import Optional

import click
from boto3 import Session

import seedfarmer.errors
import seedfarmer.mgmt.module_info as mi
from seedfarmer import DEBUG_LOGGING_FORMAT, config, enable_debug
from seedfarmer.output_utils import print_bolded
from seedfarmer.services.session_manager import SessionManager, bind_session_mgr

_logger: logging.Logger = logging.getLogger(__name__)


def _load_project() -> str:
    try:
        _logger.info("No --project provided, attempting load from seedfarmer.yaml")
        return config.PROJECT
    except FileNotFoundError:
        print_bolded("Unable to determine project to bootstrap, one of --project or a seedfarmer.yaml is required")
        raise click.ClickException("Failed to determine project identifier")


@click.group(name="remove", help="Top Level command to support removing module metadata")
def remove() -> None:
    """Remove module data"""
    pass


@remove.command(
    name="moduledata",
    help="""Remove all SSM parameters tied to the module.
        This command is meant to be run by seedfarmer ONLY!!!
        It is run within the context of the build job.
        Do not use this unless you are sure of the ramifications!
        """,
)
@click.option(
    "--deployment",
    "-d",
    type=str,
    help="The Deployment Name",
    required=True,
)
@click.option(
    "--group",
    "-g",
    type=str,
    help="The Group Name",
    required=True,
)
@click.option(
    "--module",
    "-m",
    type=str,
    help="The Module Name",
    required=True,
)
@click.option(
    "--project",
    "-p",
    help="Project identifier",
    required=False,
    default=None,
)
@click.option(
    "--profile",
    default=None,
    help="""The AWS profile used to create a session.""",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="""The AWS region used to create a session.""",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="A qualifier to use with the seedfarmer roles",
    required=False,
)
@click.option(
    "--target-account-id",
    default=None,
    help="Account Id to remove module data from, if specifed --target-region is required",
    show_default=True,
)
@click.option(
    "--target-region",
    default=None,
    help="Region to remove module data from, if specifed --target-account-id is required",
    show_default=True,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
@bind_session_mgr
def remove_module_data(
    deployment: str,
    group: str,
    module: str,
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    project: Optional[str],
    target_account_id: Optional[str],
    target_region: Optional[str],
    debug: bool,
    local: bool,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("We are removing module data for %s of group %s in %s", module, group, deployment)

    if project is None:
        project = _load_project()

    session: Session = Session(profile_name=profile, region_name=region)
    if (target_account_id is not None) != (target_region is not None):
        raise seedfarmer.errors.InvalidConfigurationError(
            "Must either specify both --target-account-id and --target-region, or neither"
        )
    elif target_account_id is not None and target_region is not None:
        session = (
            SessionManager()
            .get_or_create(project_name=project, profile=profile, region_name=region, qualifier=qualifier)
            .get_deployment_session(account_id=target_account_id, region_name=target_region)
        )
    mi.remove_module_info(deployment, group, module, session=session)



import logging
import os
from typing import Optional, Tuple
from urllib.parse import parse_qs

import git
from git import Repo

import seedfarmer.messages as messages
from seedfarmer import config
from seedfarmer.errors import InvalidConfigurationError

_logger: logging.Logger = logging.getLogger(__name__)


def get_commit_hash(repo: git.Repo) -> Optional[str]:
    try:
        return str(repo.head.object.hexsha)
    except Exception:
        _logger.warn("Could not get a commit hash, ignoring")
        return None


def clone_module_repo(git_path: str) -> Tuple[str, str, Optional[str]]:
    """Clone a git repo and return directory it is cloned into

    Rather than reinventing the wheel, we implement the Generic Git Repository functionality introduced by
    Terraform. Full documentation on the Git URL definition can be found at:
    https://www.terraform.io/language/modules/sources#generic-git-repository

    Parameters
    ----------
    git_path : str
        The Git URL specified in the Module Manifest. Full example:
        https://example.com/network.git//modules/vpc?ref=v1.2.0&depth=1

    Returns
    -------
    Tuple[str,str]
        Returns a tuple that contains (in order):
        - the full path of the seedfarmer.gitmodules where the repo was cloned to
        - the relative path to seedfarmer.gitmodules of the module code
        - the commit hash associated wtih this code
    """
    # gitpython library has started blocking non https and ssh protocols by default
    # codecommit is not _actually_ unsafe
    allow_unsafe_protocols = git_path.startswith("git::codecommit")

    git_path = git_path.replace("git::", "")
    ref: Optional[str] = None
    depth: Optional[int] = None
    module_directory = ""

    if "?" in git_path:
        git_path, query = git_path.split("?")
        query_params = parse_qs(query)
        ref = query_params.get("ref", [None])[0]
        if "depth" in query_params and query_params["depth"][0].isnumeric():
            depth = int(query_params["depth"][0])

    if ".git//" in git_path:
        git_path, module_directory = git_path.split(".git//")

    repo_directory = git_path.replace("https://", "").replace("git@", "").replace("/", "_").replace(":", "_")

    working_dir = os.path.join(
        config.OPS_ROOT, "seedfarmer.gitmodules", f"{repo_directory}_{ref.replace('/', '_')}" if ref else repo_directory
    )
    os.makedirs(working_dir, exist_ok=True)
    repo = None
    if not os.listdir(working_dir):
        if ref is not None:
            _logger.debug("Creating local repo and setting remote: %s into %s: ref=%s ", git_path, working_dir, ref)
            repo = Repo.init(working_dir)
            try:
                git.Remote.create(repo, "origin", git_path, allow_unsafe_protocols)
                repo.remotes["origin"].pull(ref, allow_unsafe_protocols=allow_unsafe_protocols)
            except git.GitError as ge:
                raise InvalidConfigurationError(f"\n Cannot Clone Repo: {ge} {messages.git_error_support()}")
        else:
            _logger.debug("Cloning %s into %s: ref=%s depth=%s", git_path, working_dir, ref, depth)
            try:
                repo = Repo.clone_from(
                    git_path, working_dir, branch=ref, depth=depth, allow_unsafe_protocols=allow_unsafe_protocols
                )
            except git.GitError as ge:
                raise InvalidConfigurationError(f"\n Cannot Clone Repo: {ge} {messages.git_error_support()}")
    else:
        _logger.debug("Pulling existing repo %s at %s: ref=%s", git_path, working_dir, ref)
        repo = Repo(working_dir)
        try:
            repo.remotes["origin"].pull(ref, allow_unsafe_protocols=allow_unsafe_protocols)
        except git.GitError as ge:
            raise InvalidConfigurationError(f"\n Cannot Clone Repo: {ge} {messages.git_error_support()}")
    commit_hash = get_commit_hash(repo)
    return (working_dir, module_directory, commit_hash)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import glob
import logging
import os
import pathlib
import shutil
import zipfile
from pprint import pformat
from typing import List, MutableSet, Optional, Tuple

from seedfarmer import CLI_ROOT
from seedfarmer.utils import create_output_dir

_logger: logging.Logger = logging.getLogger(__name__)

BUNDLE_IGNORED_FILE_PATHS: MutableSet[str] = {
    "/build/",
    "/.mypy_cache/",
    ".egg-info",
    "__pycache__",
    "/.seedfarmer.out/",
    "/dist/",
    "/node_modules/",
    "/cdk.out/",
}


def _is_valid_image_file(file_path: str) -> bool:
    return all([word not in file_path for word in BUNDLE_IGNORED_FILE_PATHS])


def _list_files(path: str) -> List[str]:
    path = os.path.join(path, "**")
    return [f for f in glob.iglob(path, recursive=True) if os.path.isfile(f) and _is_valid_image_file(file_path=f)]


def _make_zipfile(
    base_name: str, root_dir: str, base_dir: str, dry_run: bool = False, logger: Optional[logging.Logger] = None
) -> str:
    """Create a zip file from all the files under 'root_dir'/'base_dir'. Including 'base_dir' as a folder in the zip.

    The output zip file will be named 'base_name' + ".zip".  Returns the
    name of the output zip file.
    """

    zip_filename = base_name + ".zip"
    archive_dir = os.path.dirname(base_name)

    if archive_dir and not os.path.exists(archive_dir):
        if logger is not None:
            logger.info("creating %s", archive_dir)
        if not dry_run:
            os.makedirs(archive_dir)

    if logger is not None:
        logger.info("creating '%s' and adding '%s' to it", zip_filename, os.path.join(root_dir, base_dir))

    if not dry_run:
        with zipfile.ZipFile(zip_filename, "w", compression=zipfile.ZIP_DEFLATED) as zf:
            path = os.path.normpath(os.path.join(root_dir, base_dir))
            if path != os.curdir:
                zip_relative_path = os.path.relpath(pathlib.Path(path), pathlib.Path(root_dir))
                zf.write(path, zip_relative_path)
                if logger is not None:
                    logger.debug("adding '%s'", path)
            for dirpath, dirnames, filenames in os.walk(os.path.join(root_dir, base_dir)):
                for name in sorted(dirnames):
                    path = os.path.normpath(os.path.join(dirpath, name))
                    zip_relative_path = os.path.relpath(pathlib.Path(path), pathlib.Path(root_dir))
                    zf.write(path, zip_relative_path)
                    if logger is not None:
                        logger.debug("adding '%s'", path)
                for name in filenames:
                    path = os.path.normpath(os.path.join(dirpath, name))
                    if os.path.isfile(path):
                        zip_relative_path = os.path.relpath(pathlib.Path(path), pathlib.Path(root_dir))
                        zf.write(path, zip_relative_path)
                        if logger is not None:
                            logger.debug("adding '%s'", path)

    return zip_filename


def generate_dir(out_dir: str, dir: str, name: str) -> str:
    absolute_dir = os.path.realpath(dir)
    final_dir = os.path.realpath(os.path.join(out_dir, name))
    _logger.debug("absolute_dir: %s", absolute_dir)
    _logger.debug("final_dir: %s", final_dir)
    os.makedirs(final_dir, exist_ok=True)
    shutil.rmtree(final_dir)

    _logger.debug("Copying files to %s", final_dir)
    files: List[str] = _list_files(path=absolute_dir)
    if len(files) == 0:
        raise ValueError(f"{name} ({absolute_dir}) is empty!")
    for file in files:
        _logger.debug(f"***file={file}")
        relpath = os.path.relpath(file, absolute_dir)
        new_file = os.path.join(final_dir, relpath)
        _logger.debug("Copying file to %s", new_file)
        os.makedirs(os.path.dirname(new_file), exist_ok=True)
        _logger.debug("Copying file to %s", new_file)
        shutil.copy(src=file, dst=new_file)

    return final_dir


def generate_bundle(
    # fn_args: Dict[str, Any],
    dirs: Optional[List[Tuple[str, str]]] = None,
    files: Optional[List[Tuple[str, str]]] = None,
    bundle_id: Optional[str] = None,
    path_override: Optional[str] = None,
) -> str:
    bundle_dir = (
        create_output_dir(f"{bundle_id}/bundle", path_override)
        if bundle_id
        else create_output_dir("bundle", path_override)
    )
    remote_dir = os.path.dirname(bundle_dir)

    # Add the docker login script
    shutil.copy(
        src=os.path.join(CLI_ROOT, "resources/retrieve_docker_creds.py"),
        dst=os.path.join(bundle_dir, "retrieve_docker_creds.py"),
    )

    # Add the pypi credentials support
    shutil.copy(
        src=os.path.join(CLI_ROOT, "resources/pypi_mirror_support.py"),
        dst=os.path.join(bundle_dir, "pypi_mirror_support.py"),
    )

    # Add the npm credentials support
    shutil.copy(
        src=os.path.join(CLI_ROOT, "resources/npm_mirror_support.py"),
        dst=os.path.join(bundle_dir, "npm_mirror_support.py"),
    )

    _logger.debug(f"generate_bundle dirs={dirs}")
    # Extra Directories
    if dirs is not None:
        for dir, name in dirs:
            _logger.debug(f"***dir={dir}:name={name}")
            generate_dir(out_dir=bundle_dir, dir=dir, name=name)

    if files is not None:
        for src_file, name in files:
            _logger.debug(f"***file={src_file}:name={name}")
            dst_file = os.path.realpath(os.path.join(bundle_dir, name))
            if "/" in name:
                os.makedirs(os.path.dirname(dst_file), exist_ok=True)
            shutil.copy(src=src_file, dst=dst_file)

    _logger.debug("bundle_dir: %s", bundle_dir)

    files_glob = glob.glob(bundle_dir + "/**", recursive=True)
    _logger.debug("files:\n%s", pformat(files_glob))

    zip_file = _make_zipfile(base_name=bundle_dir, root_dir=remote_dir, base_dir="bundle", logger=_logger)
    return zip_file


def extract_zip(zip_path: str, extract_to: str) -> None:
    """
    Extracts a zip file to the specified directory.

    :param zip_path: Path to the .zip file
    :param extract_to: Directory to extract contents to
    """
    if not os.path.exists(zip_path):
        raise FileNotFoundError(f"Zip file not found: {zip_path}")

    os.makedirs(extract_to, exist_ok=True)

    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(extract_to)
        print(f"Extracted {zip_path} to {extract_to}")



from typing import Optional


def no_deployment_found(deployment_name: Optional[str] = None) -> str:
    msg = "We found no deployments with your active session."
    if deployment_name:
        msg = f"We found no deployments named {deployment_name} with your active session."

    checklist = """  Be sure to check:
            1. The session that you are executing the CLI from matches the ACCOUNT and REGION
            where your toolchain is configured
            2. The session has permissions to access the proper toolchain role
            """

    return msg + checklist


def git_error_support() -> str:
    return """
    1. Make sure your path to the repo is correct and valid (check your module manifests!)
    2. The credentials used to call SeedFarmer have access to the repo
    3. The credentials used to call SeedFarmer have not expired
    """



#    Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import json
import logging
from typing import Dict, Optional, Union, cast

from boto3 import Session

import seedfarmer
import seedfarmer.mgmt.bundle_support as bundle_support
from seedfarmer import config
from seedfarmer.models.deploy_responses import ModuleDeploymentResponse
from seedfarmer.models.manifests import ModuleManifest
from seedfarmer.models.transfer import ModuleDeployObject
from seedfarmer.services.session_manager import SessionManager
from seedfarmer.types.parameter_types import EnvVar
from seedfarmer.utils import generate_session_hash

_logger: logging.Logger = logging.getLogger(__name__)


class DeployModule:
    def __init__(self, mdo: ModuleDeployObject):
        self.mdo = mdo
        self.module_manifest = cast(
            ModuleManifest, mdo.deployment_manifest.get_module(str(mdo.group_name), str(mdo.module_name))
        )

    @staticmethod
    def seedfarmer_param(
        key: str, project_name: Optional[str] = None, use_project_prefix: Optional[bool] = True
    ) -> str:
        project_name = project_name if project_name else seedfarmer.config.PROJECT
        # use_project_prefix is driven by the publishGenericEnvVariables in the deployspec
        # should always be TRUE for new modules...but we are supporting legacy code
        #
        p = project_name.upper().replace("-", "_") if use_project_prefix else "SEEDFARMER"
        return f"{p}_{key}"

    def deploy_module(self) -> ModuleDeploymentResponse:
        raise NotImplementedError("Subclasses must implement 'deploy_module'")

    def destroy_module(self) -> ModuleDeploymentResponse:
        raise NotImplementedError("Subclasses must implement 'deploy_module'")

    def _prebuilt_bundle_check(self) -> Optional[str]:
        if self.mdo.seedfarmer_bucket:
            module_manifest = self.module_manifest
            account_id = str(module_manifest.get_target_account_id())
            region = str(module_manifest.target_region)
            deployment = str(self.mdo.deployment_manifest.name)
            group = self.mdo.group_name
            module = self.mdo.module_name
            bucket = self.mdo.seedfarmer_bucket
            session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)
            if bundle_support.check_bundle_exists_in_sf(deployment, str(group), str(module), bucket, session):
                return bundle_support.get_bundle_sf_path(deployment, str(group), str(module), bucket)
            else:
                return None
        else:
            return None

    def _env_vars(self, session: Optional[Session] = None) -> Dict[str, Union[str, EnvVar]]:
        use_project_prefix = not self.module_manifest.deploy_spec.publish_generic_env_variables  # type: ignore [union-attr]
        pypi_mirror_secret = (
            self.module_manifest.pypi_mirror_secret
            if self.module_manifest.pypi_mirror_secret
            else self.mdo.pypi_mirror_secret
        )
        npm_mirror_secret = (
            self.module_manifest.npm_mirror_secret
            if self.module_manifest.npm_mirror_secret
            else self.mdo.npm_mirror_secret
        )

        env_vars = (
            {
                f"{DeployModule.seedfarmer_param('PARAMETER', None, use_project_prefix)}_{p.upper_snake_case}": (
                    p.value if isinstance(p.value, str) or isinstance(p.value, EnvVar) else json.dumps(p.value)
                )
                for p in self.mdo.parameters
            }
            if self.mdo.parameters
            else {}
        )
        _logger.debug(f"use_project_prefix: {use_project_prefix}")
        _logger.debug(f"env_vars: {env_vars}")

        env_vars[DeployModule.seedfarmer_param("PROJECT_NAME", None, use_project_prefix)] = config.PROJECT
        env_vars[DeployModule.seedfarmer_param("DEPLOYMENT_NAME", None, use_project_prefix)] = str(
            self.mdo.deployment_manifest.name
        )
        env_vars[DeployModule.seedfarmer_param("MODULE_METADATA", None, use_project_prefix)] = (
            self.mdo.module_metadata if self.mdo.module_metadata is not None else ""
        )
        env_vars[DeployModule.seedfarmer_param("MODULE_NAME", None, use_project_prefix)] = (
            f"{self.mdo.group_name}-{self.module_manifest.name}"
        )
        env_vars[DeployModule.seedfarmer_param("HASH", None, use_project_prefix)] = generate_session_hash(
            session=session
        )
        if self.mdo.permissions_boundary_arn:
            env_vars[DeployModule.seedfarmer_param("PERMISSIONS_BOUNDARY_ARN", None, use_project_prefix)] = (
                self.mdo.permissions_boundary_arn
            )
        if self.mdo.docker_credentials_secret:
            env_vars["AWS_CODESEEDER_DOCKER_SECRET"] = self.mdo.docker_credentials_secret  # (LEGACY)
            env_vars["SEEDFARMER_DOCKER_SECRET"] = self.mdo.docker_credentials_secret
        if pypi_mirror_secret is not None:
            env_vars["AWS_CODESEEDER_PYPI_MIRROR_SECRET"] = str(pypi_mirror_secret)  # (LEGACY)
            env_vars["SEEDFARMER_PYPI_MIRROR_SECRET"] = str(pypi_mirror_secret)
        if npm_mirror_secret is not None:
            env_vars["AWS_CODESEEDER_NPM_MIRROR_SECRET"] = str(npm_mirror_secret)  # (LEGACY)
            env_vars["SEEDFARMER_NPM_MIRROR_SECRET"] = str(npm_mirror_secret)
        # Add the partition to env for ease of fetching
        env_vars["AWS_PARTITION"] = str(self.mdo.deployment_manifest._partition)
        env_vars["SEEDFARMER_VERSION"] = seedfarmer.__version__
        # return env_vars
        return {k: v if isinstance(v, (str, EnvVar)) else v.value for k, v in env_vars.items()}



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import logging
import os
import pathlib
from importlib.metadata import distribution
from logging import LogRecord
from typing import Any, Dict, Optional, cast

import yaml
from packaging.version import parse

import seedfarmer.errors
from seedfarmer.__metadata__ import __description__, __license__, __title__
from seedfarmer.models import ProjectSpec

_logger: logging.Logger = logging.getLogger(__name__)
__all__ = ["__description__", "__license__", "__title__"]
__version__: str = distribution(__title__).version

DEBUG_LOGGING_FORMAT = "[%(asctime)s][%(filename)-13s:%(lineno)3d] %(message)s"
INFO_LOGGING_FORMAT = "[%(asctime)s | %(levelname)s | %(threadName)s ] %(message)s"
ERROR_LOGGING_FORMAT = "[%(asctime)s | %(levelname)s | %(filename)-13s:%(lineno)3d | %(threadName)s ] %(message)s"


class LevelBasedFormatter(logging.Formatter):
    """Custom formatter that uses different formats based on log level"""

    def __init__(self) -> None:
        super().__init__()
        self.formatters = {
            logging.DEBUG: logging.Formatter(DEBUG_LOGGING_FORMAT),
            logging.INFO: logging.Formatter(INFO_LOGGING_FORMAT),
            logging.WARNING: logging.Formatter(ERROR_LOGGING_FORMAT),
            logging.ERROR: logging.Formatter(ERROR_LOGGING_FORMAT),
            logging.CRITICAL: logging.Formatter(ERROR_LOGGING_FORMAT),
        }

    def format(self, record: LogRecord) -> str:
        formatter = self.formatters.get(record.levelno, self.formatters[logging.INFO])
        return formatter.format(record)


CLI_ROOT = os.path.dirname(os.path.abspath(__file__))
DEFAULT_PROJECT_POLICY_PATH = "resources/projectpolicy.yaml"
S3_BUCKET_CFN_PATH = "resources/s3_bucket.yaml"
SEEDKIT_TEMPLATE = "resources/seedkit.template"
SEEDKIT_YAML_NAME = "seedkit.yaml"


def enable_debug(format: str) -> None:
    # For debug mode, use the custom formatter for level-based formatting
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    # Clear any existing handlers
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)

    # Create a new handler with our custom formatter
    handler = logging.StreamHandler()
    handler.setFormatter(LevelBasedFormatter())
    root_logger.addHandler(handler)

    _logger.setLevel(logging.DEBUG)
    logging.getLogger("boto3").setLevel(logging.ERROR)
    logging.getLogger("botocore").setLevel(logging.ERROR)
    logging.getLogger("s3transfer").setLevel(logging.ERROR)
    logging.getLogger("urllib3").setLevel(logging.ERROR)


def enable_info(format: str) -> None:
    # For info mode, use the custom formatter for level-based formatting
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)

    # Clear any existing handlers
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)

    # Create a new handler with our custom formatter
    handler = logging.StreamHandler()
    handler.setFormatter(LevelBasedFormatter())
    root_logger.addHandler(handler)

    _logger.setLevel(logging.INFO)
    logging.getLogger("boto3").setLevel(logging.ERROR)
    logging.getLogger("botocore").setLevel(logging.ERROR)
    logging.getLogger("s3transfer").setLevel(logging.ERROR)
    logging.getLogger("urllib3").setLevel(logging.ERROR)


enable_info(INFO_LOGGING_FORMAT)


class Config(object):
    CONFIG_FILE = "seedfarmer.yaml"
    _OPS_ROOT: Optional[str] = None

    _project_spec: Optional[ProjectSpec] = None

    def _load_config_data(self) -> None:
        count = 0
        self._OPS_ROOT = os.getcwd()
        while not os.path.exists(os.path.join(self._OPS_ROOT, self.CONFIG_FILE)):
            if count >= 4:
                _logger.error("The seedfarmer.yaml was not found at the root of your project. Please set it and rerun.")
                raise seedfarmer.errors.SeedFarmerException(
                    "The seedfarmer.yaml was not found at the root of your project. Please set it and rerun."
                )
            else:
                self._OPS_ROOT = pathlib.Path(self._OPS_ROOT).parent  # type: ignore
                count += 1

        with open(os.path.join(self._OPS_ROOT, self.CONFIG_FILE), "r") as file:
            config_data: Dict[str, Any] = yaml.safe_load(file)
            self._project_spec = ProjectSpec(**config_data)

            self._project_spec.description = (
                "NEW PROJECT" if self._project_spec.description is None else self._project_spec.description
            )
            self._project_spec.project_policy_path = (
                os.path.join(self._OPS_ROOT, self._project_spec.project_policy_path)
                if self._project_spec.project_policy_path
                else os.path.join(CLI_ROOT, DEFAULT_PROJECT_POLICY_PATH)
            )
            if self._project_spec.seedfarmer_version:
                if parse(__version__) < parse(str(self._project_spec.seedfarmer_version)):
                    msg = (
                        f"The seedfarmer.yaml specified a minimum version: "
                        f"{self._project_spec.seedfarmer_version} but you are using {__version__}"
                    )
                    raise seedfarmer.errors.SeedFarmerException(msg)

    @property
    def PROJECT(self) -> str:
        if self._project_spec is None:
            self._load_config_data()
        return str(cast(ProjectSpec, self._project_spec).project)

    @property
    def DESCRIPTION(self) -> str:
        if self._project_spec is None:
            self._load_config_data()
        return str(cast(ProjectSpec, self._project_spec).description)

    @property
    def OPS_ROOT(self) -> str:
        if self._project_spec is None:
            self._load_config_data()
        return str(self._OPS_ROOT)

    @property
    def PROJECT_POLICY_PATH(self) -> str:
        if self._project_spec is None:
            self._load_config_data()
        return str(cast(ProjectSpec, self._project_spec).project_policy_path)

    @property
    def MANIFEST_VALIDATION_FAIL_ON_UNKNOWN_FIELDS(self) -> bool:
        if self._project_spec is None:
            self._load_config_data()
        return cast(ProjectSpec, self._project_spec).manifest_validation_fail_on_unknown_fields

    @property
    def BUCKET_STORAGE_PATH(self) -> str:
        if self._project_spec is None:
            self._load_config_data()
        return str(os.path.join(CLI_ROOT, S3_BUCKET_CFN_PATH))

    @property
    def SEEDKIT_TEMPLATE_PATH(self) -> str:
        if self._project_spec is None:
            self._load_config_data()
        return str(os.path.join(CLI_ROOT, SEEDKIT_TEMPLATE))

    @property
    def SEEDKIT_YAML_FILENAME(self) -> str:
        if self._project_spec is None:
            self._load_config_data()
        return SEEDKIT_YAML_NAME


config = Config()



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import concurrent.futures
import hashlib
import json
import logging
import os
import threading
from typing import Any, Dict, List, Optional, cast

import yaml

import seedfarmer.checksum as checksum
import seedfarmer.errors
import seedfarmer.messages as messages
import seedfarmer.mgmt.archive_support as sf_archive
import seedfarmer.mgmt.deploy_utils as du
import seedfarmer.mgmt.git_support as sf_git
from seedfarmer import commands, config
from seedfarmer.commands._parameter_commands import load_parameter_values, resolve_params_for_checksum
from seedfarmer.commands._stack_commands import create_module_deployment_role, destroy_module_deployment_role
from seedfarmer.deployment.deploy_factory import DeployModuleFactory
from seedfarmer.error_handler import log_error_safely, safe_execute
from seedfarmer.mgmt.module_info import (
    get_deployspec_path,
    get_module_metadata,
    get_modulestack_path,
    remove_deployed_deployment_manifest,
    remove_deployment_manifest,
    write_deployment_manifest,
)
from seedfarmer.models import DeploySpec
from seedfarmer.models.deploy_responses import ModuleDeploymentResponse, StatusType
from seedfarmer.models.manifests import DataFile, DeploymentManifest, ModuleManifest, ModulesManifest, NetworkMapping
from seedfarmer.models.transfer import ModuleDeployObject
from seedfarmer.output_utils import (
    _print_modules,
    print_bolded,
    print_dependency_error_list,
    print_errored_modules_build_info,
    print_manifest_inventory,
    print_manifest_json,
    print_modules_build_info,
)
from seedfarmer.services import get_sts_identity_info
from seedfarmer.services._iam import get_role, get_role_arn
from seedfarmer.services.session_manager import SessionManager, bind_session_mgr
from seedfarmer.utils import get_generic_module_deployment_role_name

_logger: logging.Logger = logging.getLogger(__name__)


def _process_git_module_path(module: ModuleManifest) -> None:
    working_dir, module_directory, commit_hash = sf_git.clone_module_repo(module.path)
    module.set_local_path(os.path.join(working_dir, module_directory))
    module.commit_hash = commit_hash if commit_hash else None


def _process_archive_path(module: ModuleManifest, secret_name: Optional[str] = None) -> None:
    working_dir, module_directory = sf_archive.fetch_archived_module(module.path, secret_name)
    module.set_local_path(os.path.join(working_dir, module_directory))


def _process_data_files(
    data_files: List[DataFile],
    module_name: str,
    group_name: str,
    secret_name: Optional[str] = None,
) -> None:
    for data_file in data_files:
        if data_file.file_path.startswith("git::"):
            working_dir, module_directory, commit_hash = sf_git.clone_module_repo(data_file.file_path)
            data_file.set_local_file_path(os.path.join(working_dir, module_directory))
            data_file.set_bundle_path(module_directory)
            data_file.commit_hash = commit_hash if commit_hash else None

        elif data_file.file_path.startswith("archive::"):
            working_dir, module_directory = sf_archive.fetch_archived_module(data_file.file_path, secret_name)
            data_file.set_local_file_path(os.path.join(working_dir, module_directory))
            data_file.set_bundle_path(module_directory)

        else:
            data_file.set_local_file_path(os.path.join(config.OPS_ROOT, data_file.file_path))

    missing_files = du.validate_data_files(data_files)
    if len(missing_files) > 0:
        print(f"The following data files cannot be fetched for module {group_name}-{module_name}:")
        for missing_file in missing_files:
            print(f"  {missing_file}")
        print_bolded(message="Exiting Deployment", color="red")
        raise seedfarmer.errors.InvalidPathError("Missing DataFiles - cannot process")


def create_generic_module_deployment_role(
    account_id: str,
    region: str,
    deployment_manifest: DeploymentManifest,
) -> str:
    session = (
        SessionManager()
        .get_or_create()
        .get_deployment_session(
            account_id=account_id,
            region_name=region,
        )
    )
    role_name = get_generic_module_deployment_role_name(
        project_name=config.PROJECT,
        deployment_name=cast(str, deployment_manifest.name),
        region=region,
    )
    create_module_deployment_role(
        role_name=role_name,
        deployment_name=cast(str, deployment_manifest.name),
        permissions_boundary_arn=deployment_manifest.get_permission_boundary_arn(
            target_account=account_id,
            target_region=region,
        ),
        docker_credentials_secret=deployment_manifest.get_parameter_value(
            "dockerCredentialsSecret",
            account_alias=account_id,
            region=region,
        ),
        session=session,
        role_prefix=deployment_manifest.get_account_region_role_prefix(account_id=account_id, region=region),
    )
    return role_name


def destroy_generic_module_deployment_role(
    account_id: str,
    region: str,
    deployment_manifest: DeploymentManifest,
) -> None:
    session = (
        SessionManager()
        .get_or_create()
        .get_deployment_session(
            account_id=account_id,
            region_name=region,
        )
    )
    generic_deployment_role_name = get_generic_module_deployment_role_name(
        project_name=config.PROJECT,
        deployment_name=cast(str, deployment_manifest.name),
        region=region,
    )
    if generic_deployment_role_name:
        destroy_module_deployment_role(
            role_name=generic_deployment_role_name,
            docker_credentials_secret=deployment_manifest.get_parameter_value(
                "dockerCredentialsSecret",
                account_alias=account_id,
                region=region,
            ),
            session=session,
        )


def _execute_deploy(
    mdo: ModuleDeployObject,
) -> ModuleDeploymentResponse:
    module_manifest = cast(
        ModuleManifest, mdo.deployment_manifest.get_module(str(mdo.group_name), str(mdo.module_name))
    )
    account_id = str(module_manifest.get_target_account_id())
    region = str(module_manifest.target_region)

    mdo.parameters = load_parameter_values(
        deployment_name=cast(str, mdo.deployment_manifest.name),
        parameters=module_manifest.parameters,
        deployment_manifest=mdo.deployment_manifest,
        target_account=account_id,
        target_region=region,
    )
    module_stack_path = get_modulestack_path(str(module_manifest.get_local_path()))

    module_role_name = get_generic_module_deployment_role_name(
        project_name=config.PROJECT,
        deployment_name=cast(str, mdo.deployment_manifest.name),
        region=region,
    )
    role_prefix = mdo.deployment_manifest.get_account_region_role_prefix(account_id=account_id, region=region)

    if module_stack_path:
        _, module_role_name = commands.deploy_module_stack(
            module_stack_path=module_stack_path,
            deployment_name=cast(str, mdo.deployment_manifest.name),
            group_name=str(mdo.group_name),
            module_name=str(mdo.module_name),
            account_id=account_id,
            region=region,
            parameters=mdo.parameters,
            docker_credentials_secret=mdo.docker_credentials_secret,
            permissions_boundary_arn=mdo.permissions_boundary_arn,
            role_prefix=role_prefix,
        )

    # Get the current module's SSM if it was already loaded...
    session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)

    mdo.module_role_name = module_role_name
    mdo.module_role_arn = get_role_arn(role_name=module_role_name, session=session)

    mdo.module_metadata = json.dumps(
        get_module_metadata(
            cast(str, mdo.deployment_manifest.name), str(mdo.group_name), str(mdo.module_name), session=session
        )
    )

    if module_manifest.deploy_spec is None:
        raise seedfarmer.errors.InvalidManifestError(
            f"""Invalid value for ModuleManifest.deploy_spec in group {mdo.group_name}
              and module : {mdo.module_name}"""
        )

    (
        du.prepare_ssm_for_deploy(
            deployment_name=mdo.deployment_manifest.name,
            group_name=str(mdo.group_name),
            module_manifest=module_manifest,
            account_id=account_id,
            region=region,
        )
        if mdo.deployment_manifest.name
        else None
    )
    return DeployModuleFactory().create(mdo).deploy_module()


def _execute_destroy(mdo: ModuleDeployObject) -> Optional[ModuleDeploymentResponse]:
    module_manifest = cast(
        ModuleManifest, mdo.deployment_manifest.get_module(str(mdo.group_name), str(mdo.module_name))
    )
    if module_manifest.deploy_spec is None:
        raise seedfarmer.errors.InvalidManifestError(
            f"Invalid value for ModuleManifest.deploy_spec in group {mdo.group_name} and module : {mdo.module_name}"
        )

    target_account_id = cast(str, module_manifest.get_target_account_id())
    target_region = cast(str, module_manifest.target_region)
    role_prefix = mdo.deployment_manifest.get_account_region_role_prefix(
        account_id=target_account_id, region=target_region
    )
    session = (
        SessionManager()
        .get_or_create(role_prefix=role_prefix)
        .get_deployment_session(account_id=target_account_id, region_name=target_region)
    )
    module_metadata = get_module_metadata(
        cast(str, mdo.deployment_manifest.name), str(mdo.group_name), str(mdo.module_name), session=session
    )
    mdo.module_metadata = json.dumps(module_metadata)
    mdo.parameters = load_parameter_values(
        deployment_name=cast(str, mdo.deployment_manifest.name),
        parameters=module_manifest.parameters,
        deployment_manifest=mdo.deployment_manifest,
        target_account=target_account_id,
        target_region=target_region,
    )
    module_stack_name, module_role_name, module_stack_exists = commands.get_module_stack_info(
        deployment_name=cast(str, mdo.deployment_manifest.name),
        group_name=str(mdo.group_name),
        module_name=str(mdo.module_name),
        account_id=target_account_id,
        region=target_region,
    )

    # Use module deployment role from the metadata if it is available, otherwise fall back to default module role
    module_role_name = cast(
        str,
        module_metadata.get("ModuleDeploymentRoleName")
        if module_metadata and module_metadata.get("ModuleDeploymentRoleName")
        else module_role_name,
    )

    if get_role(role_name=module_role_name, session=session):
        mdo.module_role_name = module_role_name
    else:
        mdo.module_role_name = get_generic_module_deployment_role_name(
            project_name=config.PROJECT,
            deployment_name=cast(str, mdo.deployment_manifest.name),
            region=target_region,
        )

    if module_stack_exists:
        commands.force_manage_policy_attach(
            deployment_name=cast(str, mdo.deployment_manifest.name),
            group_name=str(mdo.group_name),
            module_name=str(mdo.module_name),
            account_id=target_account_id,
            region=target_region,
            module_role_name=mdo.module_role_name,
        )

    mdo.module_role_arn = get_role_arn(role_name=mdo.module_role_name, session=session)
    resp = DeployModuleFactory().create(mdo).destroy_module()

    if resp.status == StatusType.SUCCESS.value and module_stack_exists:
        commands.destroy_module_stack(
            cast(str, mdo.deployment_manifest.name),
            str(mdo.group_name),
            str(mdo.module_name),
            account_id=target_account_id,
            region=target_region,
            docker_credentials_secret=mdo.docker_credentials_secret,
        )

    return resp


def _deploy_validated_deployment(
    deployment_manifest: DeploymentManifest,
    deployment_manifest_wip: DeploymentManifest,
    groups_to_deploy: List[ModulesManifest],
    dryrun: bool,
) -> None:
    if groups_to_deploy:
        if dryrun:
            mods_would_deploy = [
                (_module.target_account, _module.target_region, deployment_manifest.name, _group.name, _module.name)
                for _group in groups_to_deploy
                for _module in _group.modules
            ]
            _print_modules(
                f"Modules scheduled to be deployed (created or updated): {deployment_manifest.name}", mods_would_deploy
            )
            return
        deployment_manifest_wip.groups = groups_to_deploy
        print_manifest_inventory(
            f"Modules scheduled to be deployed (created or updated): {deployment_manifest_wip.name}",
            deployment_manifest_wip,
            True,
        )
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                "DeploymentManifest for deploy after filter =  %s", json.dumps(deployment_manifest_wip.model_dump())
            )
        for _group in deployment_manifest_wip.groups:
            if len(_group.modules) > 0:
                threads = _group.concurrency if _group.concurrency else len(_group.modules)
                with concurrent.futures.ThreadPoolExecutor(max_workers=threads, thread_name_prefix="Deploy") as workers:

                    def _exec_deploy(mdo: ModuleDeployObject) -> ModuleDeploymentResponse:
                        threading.current_thread().name = (
                            f"{threading.current_thread().name}-{mdo.group_name}_{mdo.module_name}"
                        ).replace("_", "-")
                        return _execute_deploy(mdo)

                    mdos = []
                    for _module in _group.modules:
                        if _module and _module.deploy_spec:
                            mdo = ModuleDeployObject(
                                deployment_manifest=deployment_manifest_wip,
                                group_name=_group.name,
                                module_name=_module.name,
                            )
                            mdos.append(mdo)

                    deploy_response = list(workers.map(_exec_deploy, mdos))
                    _logger.debug(deploy_response)
                    (
                        print_modules_build_info("Build Info Debug Data", deploy_response)  # type: ignore
                        if _logger.isEnabledFor(logging.DEBUG)
                        else None
                    )
                    for dep_resp_object in deploy_response:
                        if dep_resp_object.status in ["ERROR", "error", "Error"]:
                            _logger.error("At least one module failed to deploy...exiting deployment")
                            print_errored_modules_build_info(
                                "These modules had errors deploying",
                                deploy_response,  # type: ignore
                            )
                            raise seedfarmer.errors.ModuleDeploymentError(
                                error_message="At least one module failed to deploy...exiting deployment"
                            )

        print_manifest_inventory(f"Modules Deployed: {deployment_manifest_wip.name}", deployment_manifest_wip, False)
    else:
        _logger.info(" All modules in %s up to date", deployment_manifest.name)
    # Write the deployment manifest once completed to preserve group order
    du.write_deployed_deployment_manifest(deployment_manifest=deployment_manifest)


def prime_target_accounts(
    deployment_manifest: DeploymentManifest,
    update_seedkit: bool = False,
    update_project_policy: bool = False,
) -> None:
    _logger.info("Priming Accounts")

    with concurrent.futures.ThreadPoolExecutor(
        max_workers=len(deployment_manifest.target_accounts_regions), thread_name_prefix="Prime-Accounts"
    ) as workers:

        def _prime_accounts(args: Dict[str, Any]) -> List[Any]:
            target_account_id = args["account_id"]
            target_region = args["region"]

            threading.current_thread().name = (
                f"{threading.current_thread().name}-{target_account_id}_{target_region}"
            ).replace("_", "-")
            _logger.info("Priming Acccount %s in %s", target_account_id, target_region)
            seedkit_stack_outputs = commands.deploy_seedkit(**args)
            seedfarmer_bucket = commands.deploy_bucket_storage_stack(**args)
            seedkit_stack_outputs["SeedfarmerArtifactBucket"] = seedfarmer_bucket
            commands.deploy_managed_policy_stack(deployment_manifest=deployment_manifest, **args)

            create_generic_module_deployment_role(
                account_id=target_account_id,
                region=target_region,
                deployment_manifest=deployment_manifest,
            )

            return [target_account_id, target_region, seedkit_stack_outputs]

        params: dict[tuple[str, str], Any] = {}
        for target_account_region in deployment_manifest.target_accounts_regions:
            account_id = target_account_region["account_id"]
            region = target_account_region["region"]
            role_prefix = target_account_region["role_prefix"]
            policy_prefix = target_account_region["policy_prefix"]
            permissions_boundary_arn = deployment_manifest.get_permission_boundary_arn(
                target_account=account_id,
                target_region=region,
            )
            param_d: dict[str, Any] = {
                "account_id": account_id,
                "region": region,
                "update_seedkit": update_seedkit,
                "update_project_policy": update_project_policy,
                "permissions_boundary_arn": permissions_boundary_arn,
            }
            if role_prefix:
                param_d["role_prefix"] = role_prefix
            if policy_prefix:
                param_d["policy_prefix"] = policy_prefix
            if target_account_region["network"] is not None:
                network = commands.load_network_values(
                    cast(NetworkMapping, target_account_region["network"]),
                    cast(Dict[str, Any], target_account_region["parameters_regional"]),
                    target_account_region["account_id"],
                    target_account_region["region"],
                )
                param_d["vpc_id"] = network.vpc_id
                param_d["private_subnet_ids"] = network.private_subnet_ids
                param_d["security_group_ids"] = network.security_group_ids

            if (account_id, region) not in params.keys():
                params[(account_id, region)] = param_d
            # param_d is a non-nested dictionary with string/bool values
            # that is acceptable to compare using equality operator
            elif params[(account_id, region)] == param_d:
                _logger.info(
                    f"Duplicate target account mapping detected for {account_id} account {region} region. Skipping..."
                )
            else:
                raise seedfarmer.errors.InvalidManifestError(
                    f"Manifest contains dulicate `targetAccountMapings` for account {account_id} region {region}, "
                    f"with different values."
                )

        output_seedkit = list(workers.map(_prime_accounts, params.values()))
        # add these to the region mappings for reference
        for out_s in output_seedkit:
            deployment_manifest.populate_metadata(
                account_id=out_s[0],
                region=out_s[1],
                seedkit_dict=out_s[2],
            )
        _logger.debug(deployment_manifest.model_dump())


def tear_down_target_accounts(deployment_manifest: DeploymentManifest, remove_seedkit: bool = False) -> None:
    # TODO: Investigate whether we need to validate the requested mappings against previously deployed mappings
    _logger.info("Tearing Down Accounts")

    with concurrent.futures.ThreadPoolExecutor(
        max_workers=len(deployment_manifest.target_accounts_regions), thread_name_prefix="Teardown-Accounts"
    ) as workers:

        def _teardown_accounts(args: Dict[str, Any]) -> None:
            target_account_id = args["account_id"]
            target_region = args["region"]
            threading.current_thread().name = (
                f"{threading.current_thread().name}-{target_account_id}_{target_region}"
            ).replace("_", "-")
            _logger.info("Tearing Down Acccount %s in %s", target_account_id, target_region)
            # Detach policies and destroy generic module deployment role
            destroy_generic_module_deployment_role(
                account_id=target_account_id,
                region=target_region,
                deployment_manifest=deployment_manifest,
            )
            commands.destroy_managed_policy_stack(**args)
            commands.destroy_bucket_storage_stack(**args)

            if remove_seedkit:
                _logger.info("Removing the seedkit tied to project %s", config.PROJECT)
                commands.destroy_seedkit(**args)

        params = [
            {
                "account_id": target_account_region["account_id"],
                "region": target_account_region["region"],
                "role_prefix": target_account_region["role_prefix"],
                "policy_prefix": target_account_region["policy_prefix"],
            }
            for target_account_region in deployment_manifest.target_accounts_regions
        ]
        _ = list(workers.map(_teardown_accounts, params))


def destroy_deployment(
    destroy_manifest: DeploymentManifest,
    remove_deploy_manifest: bool = False,
    dryrun: bool = False,
    show_manifest: bool = False,
    remove_seedkit: bool = False,
) -> None:
    """
    destroy_deployment
        Execute the destroy of a deployment based on a DeploymentManifest

    Parameters
    ----------
    deployment_manifest : DeploymentManifest
        The DeploymentManifest objec of all modules to destroy
    remove_deploy_manifest : bool, optional
        This flag indicates whether the project resource policy should be deleted.
        If there are ANY modules deployed, this should not be set to True.  This is only
        used when the entire deployment is destroyed
    dryrun : bool, optional
        This flag indicates that the DeploymentManifest object should be consumed but DOES NOT
        enact any deployment changes.

        By default False
    show_manifest : bool, optional
        This flag indicates to print out the DeploymentManifest object as s dictionary.

        By default False
    remove_seedkit: bool, optional
        This flag indicates that the project seedkit should be removed.
        This will remove it (if set to True) regardless if other deployments in the
        project use it!!  Use with caution!!

        By default False
    """
    if not destroy_manifest.groups:
        print_bolded("Nothing to destroy", "white")
        return
    print_manifest_inventory(f"Modules removed from manifest: {destroy_manifest.name}", destroy_manifest, False, "red")

    deployment_name = cast(str, destroy_manifest.name)

    print_manifest_inventory(
        f"Modules scheduled to be destroyed for: {destroy_manifest.name}", destroy_manifest, False, "red"
    )
    if not dryrun:
        for _group in reversed(destroy_manifest.groups):
            if len(_group.modules) > 0:
                threads = _group.concurrency if _group.concurrency else len(_group.modules)
                with concurrent.futures.ThreadPoolExecutor(
                    max_workers=threads, thread_name_prefix="Destroy"
                ) as workers:

                    def _exec_destroy(mdo: ModuleDeployObject) -> Optional[ModuleDeploymentResponse]:
                        threading.current_thread().name = (
                            f"{threading.current_thread().name}-{mdo.group_name}_{mdo.module_name}"
                        ).replace("_", "-")
                        return _execute_destroy(mdo)

                    mdos = []
                    for _module in _group.modules:
                        if _module.path.startswith("git::"):
                            _process_git_module_path(module=_module)
                        elif _module.path.startswith("archive::"):
                            _process_archive_path(
                                module=_module,
                                secret_name=destroy_manifest.archive_secret,
                            )

                        if _module.data_files is not None:
                            _process_data_files(
                                data_files=_module.data_files,
                                module_name=_module.name,
                                group_name=_group.name,
                                secret_name=destroy_manifest.archive_secret,
                            )

                        if _module and _module.deploy_spec:
                            mdo = ModuleDeployObject(
                                deployment_manifest=destroy_manifest, group_name=_group.name, module_name=_module.name
                            )
                            mdos.append(mdo)
                    destroy_response = list(workers.map(_exec_destroy, mdos))
                    _logger.debug(destroy_response)
                    (
                        print_modules_build_info("Build Info Debug Data", destroy_response)
                        if _logger.isEnabledFor(logging.DEBUG)
                        else None
                    )
                    for dep_resp_object in destroy_response:
                        if dep_resp_object and dep_resp_object.status in ["ERROR", "error", "Error"]:
                            _logger.error("At least one module failed to destroy...exiting deployment")
                            print_errored_modules_build_info(
                                "The following modules had errors destroying ", destroy_response
                            )
                            raise seedfarmer.errors.ModuleDeploymentError(
                                error_message="At least one module failed to destroy...exiting deployment"
                            )

        print_manifest_inventory(f"Modules Destroyed: {deployment_name}", destroy_manifest, False, "red")
        if remove_deploy_manifest:
            session = SessionManager().get_or_create().toolchain_session
            remove_deployment_manifest(deployment_name, session=session)
            remove_deployed_deployment_manifest(deployment_name, session=session)
            tear_down_target_accounts(deployment_manifest=destroy_manifest, remove_seedkit=remove_seedkit)
    if show_manifest:
        print_manifest_json(destroy_manifest)


def deploy_deployment(
    deployment_manifest: DeploymentManifest,
    module_info_index: du.ModuleInfoIndex,
    module_upstream_dep: Dict[str, List[str]],
    dryrun: bool = False,
    show_manifest: bool = False,
) -> None:
    """
    deploy_deployment
        This function takes a populated DeploymentManifest object and deploys all modules in it.
        It evaluates whether the modules have bee previously deployed and if anything has changed
        since the last deployment - ignoring if there are no changes.

    Parameters
    ----------
    deployment_manifest : DeploymentManifest
        The DeploymentManifest objec of all modules to deploy
    module_info_index:ModuleInfoIndex
        An index of all Module Info stored in SSM across all target accounts and regions
    module_upstream_dep: Dict[str, List[str]]
        A dict containing all the upstream dependencies of a module.  Each key in the dict is a module name
        with the format <group_name>-<module_name> and the value is a list of modules, each with the format
        of <group_name>-<module_name>
    dryrun : bool, optional
        This flag indicates that the DeploymentManifest object should be consumed but DOES NOT
        enact any deployment changes.
        By default False
    show_manifest : bool, optional
        This flag indicates to print out the DeploymentManifest object as s dictionary.

        By default False
    """
    deployment_manifest_wip = deployment_manifest.model_copy()
    deployment_name = cast(str, deployment_manifest_wip.name)
    _logger.debug("Setting up deployment for %s", deployment_name)

    print_manifest_inventory(
        f"Modules added to manifest: {deployment_manifest_wip.name}", deployment_manifest_wip, True
    )

    if deployment_manifest.force_dependency_redeploy:
        _logger.warn("You have configured your deployment to FORCE all dependent modules to redeploy")
        _logger.debug(f"Upstream Module Dependencies : {json.dumps(module_upstream_dep, indent=4)}")

    groups_to_deploy = []
    unchanged_modules = []
    _group_mod_to_deploy: List[str] = []
    for group in deployment_manifest_wip.groups:
        modules_to_deploy = []
        _logger.info(" Verifying all modules in %s for deploy ", group.name)
        du.validate_group_parameters(group=group)
        for module in group.modules:
            _logger.debug("Working on -- %s", module)
            if not module.path:
                raise seedfarmer.errors.InvalidManifestError("Unable to parse module manifest, `path` not specified")

            if module.path.startswith("git::"):
                _process_git_module_path(module=module)
            elif module.path.startswith("archive::"):
                _process_archive_path(
                    module=module,
                    secret_name=deployment_manifest.archive_secret,
                )

            if module.data_files is not None:
                _process_data_files(
                    data_files=module.data_files,
                    module_name=module.name,
                    group_name=group.name,
                    secret_name=deployment_manifest.archive_secret,
                )

            deployspec_path = get_deployspec_path(str(module.get_local_path()))
            with open(deployspec_path) as module_spec_file:
                module.deploy_spec = DeploySpec(**yaml.safe_load(module_spec_file))

            md5_excluded_module_files = [
                "README.md",
                "modulestack.template",
                "setup.cfg",
                "requirements-dev.txt",
                "requirements-dev.in",
                ".gitignore",
            ]

            module.bundle_md5 = checksum.get_module_md5(
                project_path=config.OPS_ROOT,
                module_path=str(module.get_local_path()),
                data_files=module.data_files,
                excluded_files=md5_excluded_module_files,
            )
            resolve_params_for_checksum(
                deployment_manifest=deployment_manifest_wip, module=module, group_name=group.name
            )

            module.manifest_md5 = hashlib.md5(
                json.dumps(module.model_dump(), sort_keys=True).encode("utf-8"),
                usedforsecurity=False,
            ).hexdigest()
            module.deployspec_md5 = hashlib.md5(open(deployspec_path, "rb").read(), usedforsecurity=False).hexdigest()

            _build_module = du.need_to_build(
                deployment_name=deployment_name,
                group_name=group.name,
                module_manifest=module,
                active_modules=_group_mod_to_deploy,
                module_upstream_dep=module_upstream_dep,
                force_redeploy_flag=cast(bool, deployment_manifest.force_dependency_redeploy),
                deployment_params_cache=module_info_index.get_module_info(
                    group=group.name,
                    account_id=cast(str, module.get_target_account_id()),
                    region=cast(str, module.target_region),
                    module_name=module.name,
                ),
            )
            if not _build_module:
                unchanged_modules.append(
                    [module.target_account, module.target_region, deployment_name, group.name, module.name]
                )
            else:
                modules_to_deploy.append(module)
                _group_mod_to_deploy.append(f"{group.name}-{module.name}")

        if modules_to_deploy:
            groups_to_deploy.append(
                ModulesManifest(
                    name=group.name, path=group.path, concurrency=group.concurrency, modules=modules_to_deploy
                )
            )
    (
        _print_modules(
            f"Modules deployed that are up to date (will not be changed): {deployment_name} ", unchanged_modules
        )
        if unchanged_modules
        else None
    )
    _deploy_validated_deployment(
        deployment_manifest=deployment_manifest,
        deployment_manifest_wip=deployment_manifest_wip,
        groups_to_deploy=groups_to_deploy,
        dryrun=dryrun,
    )
    print_bolded(f"To see all deployed modules, run seedfarmer list modules -d {deployment_name}")
    print_manifest_json(deployment_manifest) if show_manifest else None


@bind_session_mgr
@safe_execute("Deployment Apply")
def apply(
    deployment_manifest_path: str,
    profile: Optional[str] = None,
    region_name: Optional[str] = None,
    qualifier: Optional[str] = None,
    role_prefix: Optional[str] = None,
    dryrun: bool = False,
    show_manifest: bool = False,
    enable_session_timeout: bool = False,
    session_timeout_interval: int = 900,
    update_seedkit: bool = False,
    update_project_policy: bool = False,
    local: bool = False,
) -> None:
    """
    apply
        This function takes the relative path of a deployment manifest and
        generates a DeploymentManifest object necessary for deploying.  It also
        compares what is currently deployed and generates a DeploymentManifest used for destroying
        modules and groups based on what is missing from the manifest.

    Parameters
    ----------
    deployment_manifest_path : str
        Relative path to the deployment manifest
    profile : str
        If using an AWS Profile for deployment use it here
    region_name : str
        The name of the AWS region the deployment is based in for the toolchain
    qualifier : str, optional
        Any qualifier on the name of toolchain role
        Defaults to None
    role_prefix : str, optional
        IAM path prefix on the ARN of the toolchain and deployment roles
        Defaults to '/'
    dryrun : bool, optional
        This flag indicates that the deployment manifest should be consumed and a
        DeploymentManifest object be created (for both apply and destroy) but DOES NOT
        enact any deployment changes.

        By default False
    show_manifest : bool, optional
        This flag indicates to print out the DeploymentManifest object as s dictionary.

        By default False
    enable_session_timeout: bool
        If enabled, boto3 Sessions will be reset on the timeout interval
    session_timeout_interval: int
        The interval, in seconds, to reset boto3 Sessions
    update_seedkit: bool
        Force update run of seedkit, defaults to False
    update_project_policy: bool
        Force update run of managed project policy, defaults to False
    local: bool
        If set to true, use the credentials of active session and do not
        use the seedfarmer roles
        By default False

    Raises
    ------
    InvalidConfigurationError
        seedfarmer.errors.seedfarmer_errors.InvalidConfigurationError
    InvalidPathError
        seedfarmer.errors.seedfarmer_errors.InvalidPathError
    ModuleDeploymentError
        seedfarmer.errors.seedfarmer_errors.ModuleDeploymentError
    """

    manifest_path = os.path.join(config.OPS_ROOT, deployment_manifest_path)
    with open(manifest_path) as manifest_file:
        manifest_input = yaml.safe_load(manifest_file)
    ## Override all regions with the local session region if local
    if DeployModuleFactory.is_local():
        region = SessionManager().get_or_create(region_name=region_name, profile=profile).toolchain_session.region_name
        for account in manifest_input["targetAccountMappings"]:
            for mapping in account.get("regionMappings", []):
                mapping["region"] = region

    deployment_manifest = DeploymentManifest(**manifest_input)
    _logger.debug(deployment_manifest.model_dump())

    # Initialize the SessionManager for the entire project
    session_manager = SessionManager().get_or_create(
        project_name=config.PROJECT,
        profile=profile,
        qualifier=qualifier,
        role_prefix=role_prefix,
        toolchain_region=deployment_manifest.toolchain_region,
        region_name=region_name,
        enable_reaper=enable_session_timeout,
        reaper_interval=session_timeout_interval,
    )
    _, _, partition = get_sts_identity_info(session=session_manager.toolchain_session)
    deployment_manifest._partition = partition
    if not dryrun:
        write_deployment_manifest(
            cast(str, deployment_manifest.name),
            deployment_manifest.model_dump(),
            session=session_manager.toolchain_session,
        )

    for module_group in deployment_manifest.groups:
        if module_group.path and module_group.modules:
            _logger.debug("module_group: %s", module_group)
            raise seedfarmer.errors.InvalidConfigurationError(
                "Only one of the `path` or `modules` attributes can be defined on a Group"
            )
        if not module_group.path and not module_group.modules:
            _logger.debug("module_group: %s", module_group)
            raise seedfarmer.errors.InvalidConfigurationError(
                "One of the `path` or `modules` attributes must be defined on a Group"
            )
        if module_group.path:
            try:
                with open(os.path.join(config.OPS_ROOT, module_group.path)) as manifest_file:
                    if DeployModuleFactory.is_local():
                        region = session_manager.toolchain_session.region_name
                        updated_manifests = []
                        for m in yaml.safe_load_all(manifest_file):
                            m["targetRegion"] = region
                            updated_manifests.append(ModuleManifest(**m))
                        module_group.modules = updated_manifests
                    else:
                        module_group.modules = [ModuleManifest(**m) for m in yaml.safe_load_all(manifest_file)]
            except FileNotFoundError as fe:
                log_error_safely(
                    _logger, fe, f"Cannot parse a file at {os.path.join(config.OPS_ROOT, module_group.path)}"
                )
                _logger.error("Verify (in deployment manifest) that relative path to the module manifest is correct")
                raise seedfarmer.errors.InvalidPathError(f"Cannot parse manifest file path at {module_group.path}")
            except Exception as e:
                log_error_safely(_logger, e, "Failed to parse module manifest file")
                _logger.error(f"Module manifest file parsing error: {e}")
                _logger.error("Verify that YAML syntax is valid and all required fields are present")
                raise seedfarmer.errors.InvalidManifestError(f"Cannot parse manifest file: {e}")
    deployment_manifest.validate_and_set_module_defaults()

    prime_target_accounts(
        deployment_manifest=deployment_manifest,
        update_seedkit=update_seedkit,
        update_project_policy=update_project_policy,
    )

    module_info_index = du.populate_module_info_index(deployment_manifest=deployment_manifest)
    destroy_manifest = du.filter_deploy_destroy(deployment_manifest, module_info_index)

    module_depends_on_dict, module_dependencies_dict = du.generate_dependency_maps(manifest=deployment_manifest)
    _logger.debug("module_depends_on_dict: %s", json.dumps(module_depends_on_dict))
    _logger.debug("module_dependencies_dict: %s", json.dumps(module_dependencies_dict))
    violations = du.validate_module_dependencies(module_dependencies_dict, destroy_manifest)
    if violations:
        print_dependency_error_list(
            header_message="The following modules requested for destroy have dependencies that prevent destruction:",
            errored_list=violations,
        )
        raise seedfarmer.errors.InvalidConfigurationError("Modules cannot be destroyed due to dependencies")

    destroy_deployment(
        destroy_manifest=destroy_manifest,
        remove_deploy_manifest=False,
        dryrun=dryrun,
        show_manifest=show_manifest,
    )
    deploy_deployment(
        deployment_manifest=deployment_manifest,
        module_info_index=module_info_index,
        module_upstream_dep=module_depends_on_dict,
        dryrun=dryrun,
        show_manifest=show_manifest,
    )


@bind_session_mgr
@safe_execute("Deployment Destroy")
def destroy(
    deployment_name: str,
    profile: Optional[str] = None,
    region_name: Optional[str] = None,
    qualifier: Optional[str] = None,
    role_prefix: Optional[str] = None,
    dryrun: bool = False,
    show_manifest: bool = False,
    remove_seedkit: bool = False,
    enable_session_timeout: bool = False,
    session_timeout_interval: int = 900,
    local: bool = False,
) -> None:
    """
    destroy
        This function takes the name of a deployment and destroy all artifacts related.

    Parameters
    ----------
    deployment_name : str
        The name of the deployment to destroy
    profile : str
        If using an AWS Profile for deployment use it here
    region_name : str
        The name of the AWS region the deployment is based in for the toolchain
    qualifier : str, optional
        Any qualifier on the name of toolchain role
        Defaults to None
    role_prefix : str, optional
        IAM path prefix on the ARN of the toolchain and deployment roles
        Defaults to '/'
    dryrun : bool, optional
        This flag indicates that the deployment WILL NOT
        enact any deployment changes.
        By default False
    remove_seedkit: bool, optional
        This flag indicates that the project seedkit should be removed.
        This will remove it (if set to True) regardless if other deployments in the
        project use it!!  Use with caution!!

        By default False
    show_manifest : bool, optional
        This flag indicates to print out the DeploymentManifest object as s dictionary.

        By default False
    enable_session_timeout: bool
        If enabled, boto3 Sessions will be reset on the timeout interval
    session_timeout_interval: int
        The interval, in seconds, to reset boto3 Sessions
    local: bool
        If set to true, use the credentials of active session and do not
        use the seedfarmer roles
        By default False
    Raises
    ------
    InvalidConfigurationError
        seedfarmer.errors.seedfarmer_errors.InvalidConfigurationError
    InvalidPathError
        seedfarmer.errors.seedfarmer_errors.InvalidPathError
    ModuleDeploymentError
        seedfarmer.errors.seedfarmer_errors.ModuleDeploymentError
    """
    project = config.PROJECT
    _logger.debug("Preparing to destroy %s", deployment_name)

    session_manager = SessionManager().get_or_create(
        project_name=project,
        profile=profile,
        region_name=region_name,
        qualifier=qualifier,
        role_prefix=role_prefix,
        enable_reaper=enable_session_timeout,
        reaper_interval=session_timeout_interval,
    )
    destroy_manifest = du.generate_deployed_manifest(deployment_name=deployment_name, skip_deploy_spec=False)
    if destroy_manifest:
        _, _, partition = get_sts_identity_info(session=session_manager.toolchain_session)
        destroy_manifest._partition = partition
        destroy_manifest.validate_and_set_module_defaults()
        prime_target_accounts(
            deployment_manifest=destroy_manifest,
            update_seedkit=False,
            update_project_policy=False,
        )
        destroy_deployment(
            destroy_manifest,
            remove_deploy_manifest=True,
            dryrun=dryrun,
            show_manifest=show_manifest,
            remove_seedkit=remove_seedkit,
        )
    else:
        account_id, _, _ = get_sts_identity_info(session=session_manager.toolchain_session)
        region = session_manager.toolchain_session.region_name
        _logger.info(
            """Deployment %s was not found in project %s in account %s and region %s
                     """,
            deployment_name,
            project,
            account_id,
            region,
        )
        print_bolded(message=messages.no_deployment_found(deployment_name=deployment_name), color="yellow")



#!/usr/bin/env python

import argparse
import json
import logging
import os
import subprocess
from typing import Dict, Optional, Tuple, cast

import boto3
from botocore.exceptions import ClientError

logger = logging.getLogger()
logger.setLevel(logging.INFO)


def get_secret(secret_name: str) -> Dict[str, Dict[str, str]]:
    region_name = os.environ.get("AWS_DEFAULT_REGION")

    try:
        session = boto3.session.Session()
        client = session.client(service_name="secretsmanager", region_name=region_name)
        get_secret_value_response = client.get_secret_value(SecretId=secret_name)
    except ClientError as ce:
        logger.info("Secret with SecretId '%s' could not be retrieved from SecretsManager", secret_name)
        print(f"Secret with SecretId {secret_name} could not be retrieved from SecretsManager - {ce}")
        exit(1)
    except FileNotFoundError:
        logger.info("Make sure AWW credentials are set")
        print("Make sure credentials  AWW credentials are set")
        exit(1)
    else:
        return cast(Dict[str, Dict[str, str]], json.loads(get_secret_value_response.get("SecretString", "{}")))


def create_url(url: str, username: Optional[str] = None, password: Optional[str] = None) -> Tuple[str, str]:
    if url.startswith("http://"):
        logger.info("URL is unsecured, no passwords will be set and url used as-is")
        return url, url
    if url.startswith("https://"):
        if username is not None and password is not None:
            s = url.split("https://")
            secured_url = f"https://{username}:{password}@{s[1]}"
            obfusctated_url = f"https://XXXX:XXXX@{s[1]}"
            return secured_url, obfusctated_url
        else:
            logger.info("Username and / or password not set, using URL as-is")
            return url, url
    return url, url


def setup_pypi(obfusctated_url: str, secured_url: str) -> None:
    logger.info("Calling pip config with %s", obfusctated_url)
    print(f"pip config being set for : {obfusctated_url}")
    subprocess.call(["pip", "config", "set", "global.index-url", secured_url])


def setup_uv(obfusctated_url: str, secured_url: str) -> None:
    logger.info("Setting up uv.config %s", obfusctated_url)
    print(f"uv configured at  ~/.config/uv/uv.toml : {obfusctated_url}")
    toml_content = f"""[[index]]
    name = "internal-proxy"
    url = "{secured_url}"
    default = true
    """
    uv_config_path = os.path.expanduser("~/.config/uv")
    os.makedirs(uv_config_path, exist_ok=True)

    # Write to file
    with open(os.path.join(uv_config_path, "uv.toml"), "w") as f:
        f.write(toml_content)


def main(url: str) -> None:
    secret_name = os.environ.get("SEEDFARMER_PYPI_MIRROR_SECRET", "NO_SECRET")
    # Backwards Compatibility
    if secret_name == "NO_SECRET":
        secret_name = os.environ.get("AWS_CODESEEDER_PYPI_MIRROR_SECRET", "NO_SECRET")
    elif secret_name == "NO_SECRET":
        secret_name = os.environ.get("AWS_CODESEEDER_MIRROR_SECRET", "NO_SECRET")
    username = None
    password = None
    if secret_name not in ["NO_SECRET"]:
        secret_name_key = secret_name.split("::")[0] if "::" in secret_name else secret_name
        key = secret_name.split("::")[1] if "::" in secret_name else "pypi"
        creds = get_secret(secret_name=secret_name_key)
        if key in creds.keys():
            username = creds[key]["username"] if creds[key].get("username") else None
            password = creds[key]["password"] if creds[key].get("password") else None

    secured_url, obfusctated_url = create_url(url, username, password)
    setup_pypi(obfusctated_url=obfusctated_url, secured_url=secured_url)
    setup_uv(obfusctated_url=obfusctated_url, secured_url=secured_url)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="URL to evaluate")
    parser.add_argument("url", type=str, help="The url to set in pypi'")
    args = parser.parse_args()
    main(args.url)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Any, Dict, List, Optional

from rich.console import Console
from rich.table import Table

from seedfarmer.models.deploy_responses import ModuleDeploymentResponse
from seedfarmer.models.manifests import DeploymentManifest

console = Console(record=True)


def print_deployment_inventory(description: str, dep: List[str], color: str = "yellow") -> None:
    """
    A helper method to print the list of deployments in a table to console

    Parameters
    ----------
    description : str
        Give the table a title
    dep : List[str]
        A list of strings of the deployment names
    color : str, optional
        The color of the Title , by default "yellow"
    """
    table = Table(title=f"[bold {color}]{description}", title_justify="left")

    table.add_column("Deployment", justify="left", style="cyan", no_wrap=True)
    for deployment in dep:
        table.add_row(deployment)

    console.print(table)


def print_manifest_inventory(
    description: str, dep: DeploymentManifest, show_path: bool = False, color: str = "yellow"
) -> None:
    """
    A helper method to print the content of a DeploymentManifest object in a table to console

    Parameters
    ----------
    description : str
        Give the table a title
    dep : DeploymentManifest
        A populated Deplopyment Manifest object
    show_path : bool, optional
        Show the relative path of the module code, by default False
    color : str, optional
        The color of the Title, by default "yellow"
    """
    table = Table(title=f"[bold {color}]{description}", title_justify="left")

    table.add_column("Account", justify="left", style="yellow", no_wrap=True)
    table.add_column("Region", justify="left", style="white", no_wrap=True)
    table.add_column("Deployment", justify="left", style="cyan", no_wrap=True)
    table.add_column("Group", justify="left", style="magenta")
    table.add_column("Module", justify="left", style="green")
    table.add_column("Path", justify="left") if show_path else None
    d_name = dep.name
    for group in dep.groups:
        g_name = group.name
        for module in group.modules:
            m_name = module.name
            m_path = module.path
            region = module.target_region
            account = module.target_account
            if show_path:
                table.add_row(account, region, d_name, g_name, m_name, m_path)
            else:
                table.add_row(account, region, d_name, g_name, m_name)

    console.print(table)


def _print_modules(description: str, modules_list: List[Any]) -> None:
    table = Table(title=f"[bold yellow]{description}", title_justify="left")

    table.add_column("Account", justify="left", style="yellow", no_wrap=True)
    table.add_column("Region", justify="left", style="white", no_wrap=True)
    table.add_column("Deployment", style="cyan", no_wrap=True)
    table.add_column("Group", justify="left", style="magenta")
    table.add_column("Module", justify="left", style="green")
    for lst in modules_list:
        table.add_row(lst[0], lst[1], lst[2], lst[3], lst[4])

    console.print(table)


def print_manifest_json(dep: DeploymentManifest) -> None:
    """
    Pretty-print to console a json representation of the DeploymentManifest

    Parameters
    ----------
    dep : DeploymentManifest
        The DeploymentManifest to be printed as json
    """
    console.print(dep.dict(), overflow="ignore", crop=False)


def print_json(payload: Optional[Dict[str, Any]]) -> None:
    """
    Pretty-print to console a json representation of the DeploymentManifest

    Parameters
    ----------
    dep : DeploymentManifest
        The DeploymentManifest to be printed as json
    """
    console.print(payload, overflow="ignore", crop=False)


def print_bolded(message: str, color: str = "yellow") -> None:
    """
    Print a string message to console in a requested color and bolded

    Parameters
    ----------
    message : str
        The String
    color : str, optional
        The color you want the message printed in.... default is "yellow"
    """
    console.print(f"[bold {color}]{message}")


def print_modules_build_info(
    description: str, modules_data: List[Optional[ModuleDeploymentResponse]], color: str = "white"
) -> None:
    """
    Print the modules that have errored on deployment of a group

    Parameters
    ----------
    description : str
        The custom text
    modules_data : List[Optional[ModuleDeploymentResponse]]
        The object containing the metadata related to the failure
    color : str, optional
        The color of the description text, by default "white"
    """
    console.print(f"[bold {color}]{description}", crop=False)
    for r_obj in modules_data:
        console.print(
            f" [cyan]{r_obj.deployment}-{r_obj.group}-{r_obj.module} status: {r_obj.status}",  # type: ignore
            crop=False,
        )
        (
            console.print(f"    {r_obj.codebuild_metadata.build_url}", crop=False)  # type: ignore
            if r_obj.codebuild_metadata and r_obj.codebuild_metadata.build_url  # type: ignore
            else None
        )


def print_errored_modules_build_info(
    description: str, modules_data: List[Optional[ModuleDeploymentResponse]], color: str = "red"
) -> None:
    """
    Print the modules that have errored on deployment of a group

    Parameters
    ----------
    description : str
        The custom text
    modules_data : List[Optional[ModuleDeploymentResponse]]
        The object containing the metadata related to the failure
    color : str, optional
        The color of the description text, by default "red"
    """
    console.print(f"[bold {color}]{description}", crop=False)
    for r_obj in modules_data:
        if r_obj and r_obj.status in ["ERROR", "error", "Error"]:
            console.print(f" [cyan]{r_obj.deployment}-{r_obj.group}-{r_obj.module}", crop=False)
            (
                console.print(f"    {r_obj.codebuild_metadata.build_url}", crop=False)
                if r_obj.codebuild_metadata and r_obj.codebuild_metadata.build_url
                else None
            )


def print_dependency_error_list(header_message: str, errored_list: List[Dict[str, List[str]]]) -> None:
    """
    Print out a list of error strings

    Parameters
    ----------
    header_message : str
        The header message printed first
    errored_list : List[Dict[str, List[str]]]
        A list of dictionaries of lists to print
    """
    console.print(f"[bold yellow] {header_message}")
    for entry in errored_list:
        for mod in entry.keys():
            console.print(f"  [bold cyan]{mod} ")
            console.print(f"    [cyan]{entry.get(mod)}")


def print_dependency_list(header_message: str, modules: List[str]) -> None:
    """
    Prints the list of modules

    Parameters
    ----------
    header_message : str
         The header message printed first
    modules : List[str]
        The module names in a list
    """
    console.print(f"[bold yellow] {header_message}")
    console.print(f"  [cyan]{modules} ")



#    Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import os
from typing import Dict, List, Optional, Union, cast

import seedfarmer
import seedfarmer.deployment.codebuild_local as codebuild_local
import seedfarmer.mgmt.bundle as bundle
import seedfarmer.services._codebuild as codebuild
from seedfarmer import config
from seedfarmer.commands._runtimes import get_runtimes
from seedfarmer.deployment.deploy_base import DeployModule
from seedfarmer.errors import InvalidConfigurationError
from seedfarmer.models.deploy_responses import ModuleDeploymentResponse, StatusType
from seedfarmer.models.manifests import ModuleManifest
from seedfarmer.types.parameter_types import EnvVar
from seedfarmer.utils import LiteralStr, apply_literalstr, create_output_dir, register_literal_str


class DeployLocalModule(DeployModule):
    def _codebuild_install_commands(
        self,
        module_manifest: ModuleManifest,
        stack_outputs: Optional[Dict[str, str]],
        runtimes: Optional[Dict[str, str]] = None,
    ) -> List[Union[str, LiteralStr]]:
        python_version = runtimes.get("python", "3.11") if runtimes else "3.11"
        npm_mirror = module_manifest.npm_mirror if module_manifest.npm_mirror is not None else self.mdo.npm_mirror
        pypi_mirror = module_manifest.pypi_mirror if module_manifest.pypi_mirror is not None else self.mdo.pypi_mirror

        install = []
        install.append("mkdir -p /var/scripts/")
        install.append(
            "mv $CODEBUILD_SRC_DIR/bundle/retrieve_docker_creds.py /var/scripts/retrieve_docker_creds.py || true"
        )
        install.append(
            "/var/scripts/retrieve_docker_creds.py && echo 'Docker logins successful' || echo 'Docker logins failed'"
        )

        if pypi_mirror is not None:
            install.append("mv $CODEBUILD_SRC_DIR/bundle/pypi_mirror_support.py /var/scripts/pypi_mirror_support.py")
            install.append(f"/var/scripts/pypi_mirror_support.py {pypi_mirror} && echo 'Pypi Mirror Set'")

        if npm_mirror:
            install.append("mv $CODEBUILD_SRC_DIR/bundle/npm_mirror_support.py /var/scripts/npm_mirror_support.py")
            install.append(f"/var/scripts/npm_mirror_support.py {npm_mirror} && echo 'NPM Mirror Set'")

        if stack_outputs and "CodeArtifactDomain" in stack_outputs and "CodeArtifactRepository" in stack_outputs:
            ca_domain = stack_outputs["CodeArtifactDomain"]
            ca_repository = stack_outputs["CodeArtifactRepository"]
            region = str(module_manifest.target_region)
            account_id = str(module_manifest.get_target_account_id())
            install.append("mkdir -p ~/.config/uv")
            install.append(
                apply_literalstr(
                    "cat <<EOF > ~/.config/uv/uv.toml\n"
                    "[[index]]\n"
                    'name = "codeartifact"\n'
                    f'url = "https://{ca_domain}-{account_id}.d.codeartifact.{region}.amazonaws.com/pypi/{ca_repository}/simple/"\n'
                    "EOF"
                )
            )

            install.append(
                apply_literalstr(
                    (
                        f"aws codeartifact login --tool pip --domain {ca_domain} "
                        f"--repository {ca_repository} --region {region}"
                    )
                )
            )
            install.append(
                apply_literalstr(
                    "export CODEARTIFACT_AUTH_TOKEN=$(aws codeartifact get-authorization-token "
                    f"--domain {ca_domain} "
                    f"--domain-owner {account_id} "
                    f"--region {region} "
                    "--query authorizationToken "
                    "--output text)"
                )
            )
            install.append("export UV_INDEX_CODEARTIFACT_USERNAME=aws")
            install.append('export UV_INDEX_CODEARTIFACT_PASSWORD="$CODEARTIFACT_AUTH_TOKEN"')

        # needed to make sure both the tool and lib are accessible in the venv
        install.append("pip install uv --disable-pip-version-check --quiet --root-user-action=ignore")
        install.append("export PATH=$PATH:/root/.local/bin")
        install.append(f"uv venv ~/.venv --python {python_version} --seed --quiet")
        install.append(f"uv tool install seed-farmer=={seedfarmer.__version__} --quiet")

        return install

    def deploy_module(self) -> ModuleDeploymentResponse:
        deployment_manifest = self.mdo.deployment_manifest
        group = self.mdo.group_name
        deployment_manifest = self.mdo.deployment_manifest
        module_manifest = cast(
            ModuleManifest, self.mdo.deployment_manifest.get_module(str(self.mdo.group_name), str(self.mdo.module_name))
        )
        use_project_prefix = not module_manifest.deploy_spec.publish_generic_env_variables  # type: ignore [union-attr]
        yaml = register_literal_str()

        if module_manifest.deploy_spec is None or module_manifest.deploy_spec.deploy is None:
            raise InvalidConfigurationError("Missing `deploy` in module's deployspec.yaml")

        account_id = str(module_manifest.get_target_account_id())
        region = str(module_manifest.target_region)
        env_vars = self._env_vars()

        metadata_env_var = DeployModule.seedfarmer_param("MODULE_METADATA", None, use_project_prefix)

        metadata_put = [
            f"if [[ -f {metadata_env_var} ]]; then export {metadata_env_var}=$(cat {metadata_env_var}); fi",
            (
                f"echo ${metadata_env_var} | seedfarmer store moduledata "
                f"-d {deployment_manifest.name} -g {group} -m {module_manifest.name} --region {region} --local"
            ),
        ]

        md5_put = [
            (
                f"echo {module_manifest.bundle_md5} | seedfarmer store md5 -d {self.mdo.deployment_manifest.name} "
                f"-g {self.mdo.group_name} -m {module_manifest.name} -t bundle --region {region} --local"
            )
        ]

        metadata_env_var = DeployModule.seedfarmer_param("MODULE_METADATA", None, use_project_prefix)
        sf_version_add = [f"seedfarmer metadata add -k SeedFarmerDeployed -v {seedfarmer.__version__} || true"]
        module_role_name_add = [
            f"seedfarmer metadata add -k ModuleDeploymentRoleName -v {self.mdo.module_role_name} || true"
        ]
        githash_add = (
            [f"seedfarmer metadata add -k SeedFarmerModuleCommitHash -v {module_manifest.commit_hash} || true"]
            if module_manifest.commit_hash
            else []
        )

        extra_file_bundle = {config.CONFIG_FILE: os.path.join(config.OPS_ROOT, config.CONFIG_FILE)}
        module_path = os.path.join(config.OPS_ROOT, str(module_manifest.get_local_path()))
        dirs = {"module": module_path}
        dirs_tuples = [(v, k) for k, v in dirs.items()]

        extra_files = {}
        if module_manifest.data_files is not None:
            extra_files = {
                f"module/{data_file.get_bundle_path()}": data_file.get_local_file_path()
                for data_file in module_manifest.data_files
            }
        if extra_files is not None:
            extra_file_bundle.update(extra_files)  # type: ignore [arg-type]

        files_tuples = [(v, f"{k}") for k, v in extra_file_bundle.items()]

        _phases = module_manifest.deploy_spec.deploy.phases
        codebuild_image = (
            module_manifest.codebuild_image if module_manifest.codebuild_image is not None else self.mdo.codebuild_image
        )
        codebuild_image = (
            codebuild_image if codebuild_image else "public.ecr.aws/codebuild/amazonlinux2-x86_64-standard:5.0"
        )

        if codebuild_image.startswith("aws/codebuild/"):
            codebuild_image = f"public.ecr.{codebuild_image}"

        bundle_id = f"{self.mdo.deployment_manifest.name}-{self.mdo.group_name}-{module_manifest.name}"
        output_override = f".seedfarmerlocal-{bundle_id}"

        local_path = create_output_dir(f"{bundle_id}", output_override)
        bundle.generate_bundle(dirs=dirs_tuples, files=files_tuples, bundle_id=bundle_id, path_override=output_override)
        stack_outputs = deployment_manifest.get_region_seedfarmer_metadata(account_id=account_id, region=region)

        runtime_versions = get_runtimes(codebuild_image=codebuild_image, runtime_overrides=self.mdo.runtime_overrides)
        install_commands = self._codebuild_install_commands(module_manifest, stack_outputs, runtime_versions)

        buildspec = codebuild.generate_spec(
            cmds_install=install_commands
            + [". ~/.venv/bin/activate"]
            + ["cd ${CODEBUILD_SRC_DIR}/bundle"]
            + ["cd module/"]
            + _phases.install.commands,
            cmds_pre=[". ~/.venv/bin/activate"]
            + ["cd ${CODEBUILD_SRC_DIR}/bundle"]
            + ["cd module/"]
            + _phases.pre_build.commands,
            cmds_build=[". ~/.venv/bin/activate"]
            + ["cd ${CODEBUILD_SRC_DIR}/bundle"]
            + ["cd module/"]
            + _phases.build.commands,
            cmds_post=[". ~/.venv/bin/activate"]
            + ["cd ${CODEBUILD_SRC_DIR}/bundle"]
            + ["cd module/"]
            + _phases.post_build.commands
            + md5_put
            + sf_version_add
            + module_role_name_add
            + githash_add
            + metadata_put
            + ["cd ${CODEBUILD_SRC_DIR}"]
            + ["chmod -R 777 ${CODEBUILD_SRC_DIR}"],  # makes sure output isn't locked
            abort_phases_on_failure=True,
            runtime_versions=runtime_versions,
        )

        buildspec_dir = create_output_dir(f"{bundle_id}/buildspec", output_override)
        with open(os.path.join(buildspec_dir, "buildspec.yaml"), "w") as file:
            # file.write(yaml.dump(buildspec))
            yaml.dump(buildspec, file)

        build_info = codebuild_local.run(
            local_path, {k: v.value if isinstance(v, EnvVar) else v for k, v in env_vars.items()}, codebuild_image
        )

        return ModuleDeploymentResponse(
            deployment=self.mdo.deployment_manifest.name,
            group=self.mdo.group_name,
            module=module_manifest.name,
            status=StatusType.SUCCESS.value if build_info.status.value in ["SUCCEEDED"] else StatusType.ERROR.value,
        )

    def destroy_module(self) -> ModuleDeploymentResponse:
        deployment_name = self.mdo.deployment_manifest.name
        group_name = self.mdo.group_name
        module_name = self.mdo.module_name

        module_manifest = cast(
            ModuleManifest, self.mdo.deployment_manifest.get_module(str(group_name), str(module_name))
        )
        use_project_prefix = not module_manifest.deploy_spec.publish_generic_env_variables  # type: ignore [union-attr]
        yaml = register_literal_str()

        if module_manifest.deploy_spec is None or module_manifest.deploy_spec.destroy is None:
            raise seedfarmer.errors.InvalidConfigurationError(
                f"Missing `destroy` in module: {module_manifest.name} with deployspec.yaml"
            )

        account_id = str(module_manifest.get_target_account_id())
        region = str(module_manifest.target_region)
        env_vars = self._env_vars()

        metadata_env_var = DeployModule.seedfarmer_param("MODULE_METADATA", None, use_project_prefix)

        metadata_put = [
            f"if [[ -f {metadata_env_var} ]]; then export {metadata_env_var}=$(cat {metadata_env_var}); fi",
        ]
        remove_ssm = [
            (
                f"seedfarmer remove moduledata "
                f"-d {deployment_name} -g {group_name} -m {module_manifest.name} --region {region} --local"
            )
        ]
        export_info = [
            f"export DEPLOYMENT={self.mdo.deployment_manifest.name}",
            f"export GROUP={self.mdo.group_name}",
            f"export MODULE={module_manifest.name}",
        ]

        extra_file_bundle = {config.CONFIG_FILE: os.path.join(config.OPS_ROOT, config.CONFIG_FILE)}
        module_path = os.path.join(config.OPS_ROOT, str(module_manifest.get_local_path()))
        dirs = {"module": module_path}
        dirs_tuples = [(v, k) for k, v in dirs.items()]
        extra_files = {}
        if module_manifest.data_files is not None:
            extra_files = {
                f"module/{data_file.get_bundle_path()}": data_file.get_local_file_path()
                for data_file in module_manifest.data_files
            }
        if extra_files is not None:
            extra_file_bundle.update(extra_files)  # type: ignore [arg-type]
        files_tuples = [(v, f"{k}") for k, v in extra_file_bundle.items()]

        codebuild_image = (
            module_manifest.codebuild_image if module_manifest.codebuild_image is not None else self.mdo.codebuild_image
        )
        codebuild_image = (
            codebuild_image if codebuild_image else "public.ecr.aws/codebuild/amazonlinux2-x86_64-standard:5.0"
        )

        if codebuild_image.startswith("aws/codebuild/"):
            codebuild_image = f"public.ecr.{codebuild_image}"

        bundle_id = f"{deployment_name}-{group_name}-{module_name}"
        output_override = f".seedfarmerlocal-{bundle_id}"

        local_path = create_output_dir(f"{bundle_id}", output_override)
        bundle.generate_bundle(dirs=dirs_tuples, files=files_tuples, bundle_id=bundle_id, path_override=output_override)

        _phases = module_manifest.deploy_spec.destroy.phases
        stack_outputs = self.mdo.deployment_manifest.get_region_seedfarmer_metadata(
            account_id=account_id, region=region
        )

        runtime_versions = get_runtimes(codebuild_image=codebuild_image, runtime_overrides=self.mdo.runtime_overrides)
        install_commands = self._codebuild_install_commands(module_manifest, stack_outputs, runtime_versions)

        buildspec = codebuild.generate_spec(
            cmds_install=install_commands
            + [". ~/.venv/bin/activate"]
            + ["cd ${CODEBUILD_SRC_DIR}/bundle"]
            + ["cd module/"]
            + _phases.install.commands,
            cmds_pre=[". ~/.venv/bin/activate"]
            + ["cd ${CODEBUILD_SRC_DIR}/bundle"]
            + ["cd module/"]
            + _phases.pre_build.commands
            + export_info,
            cmds_build=[". ~/.venv/bin/activate"]
            + ["cd ${CODEBUILD_SRC_DIR}/bundle"]
            + ["cd module/"]
            + _phases.build.commands,
            cmds_post=[". ~/.venv/bin/activate"]
            + ["cd ${CODEBUILD_SRC_DIR}/bundle"]
            + ["cd module/"]
            + _phases.post_build.commands
            + metadata_put
            + remove_ssm
            + ["cd ${CODEBUILD_SRC_DIR}"]
            + ["chmod -R 777 ${CODEBUILD_SRC_DIR}"],  # makes sure output isn't locked
            abort_phases_on_failure=True,
            runtime_versions=runtime_versions,
        )

        buildspec_dir = create_output_dir(f"{bundle_id}/buildspec", output_override)
        with open(os.path.join(buildspec_dir, "buildspec.yaml"), "w") as file:
            yaml.dump(buildspec, file)

        build_info = codebuild_local.run(
            local_path, {k: v.value if isinstance(v, EnvVar) else v for k, v in env_vars.items()}, codebuild_image
        )

        return ModuleDeploymentResponse(
            deployment=deployment_name,
            group=group_name,
            module=module_name,
            status=StatusType.SUCCESS.value if build_info.status.value in ["SUCCEEDED"] else StatusType.ERROR.value,
        )



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import random
import time
from typing import TYPE_CHECKING, Any, Callable, Dict, Literal, Optional, Tuple, Union, cast, overload

import boto3
import botocore.config
import botocore.exceptions
from boto3 import Session
from botocore.auth import SigV4Auth
from botocore.awsrequest import AWSRequest
from botocore.credentials import Credentials

import seedfarmer
import seedfarmer.errors

if TYPE_CHECKING:
    from boto3.resources.base import ServiceResource
    from botocore.client import BaseClient
    from mypy_boto3_cloudformation.client import CloudFormationClient
    from mypy_boto3_codebuild import CodeBuildClient
    from mypy_boto3_iam import IAMClient, IAMServiceResource
    from mypy_boto3_logs.client import CloudWatchLogsClient
    from mypy_boto3_s3 import S3Client, S3ServiceResource
    from mypy_boto3_secretsmanager import SecretsManagerClient
    from mypy_boto3_ssm import SSMClient
    from mypy_boto3_sts.client import STSClient

_logger: logging.Logger = logging.getLogger(__name__)


def setup_proxies() -> Dict[str, Optional[str]]:
    proxies = {}
    proxies["http"] = os.getenv("HTTP_PROXY", None)
    proxies["https"] = os.getenv("HTTPS_PROXY", None)
    _logger.debug("Proxies Configured: %s", proxies)
    return proxies


def get_botocore_config() -> botocore.config.Config:
    return botocore.config.Config(
        retries={"max_attempts": 5},
        connect_timeout=10,
        max_pool_connections=10,
        user_agent_extra=f"seedfarmer/{seedfarmer.__version__} seedfarmer/project/{seedfarmer.config.PROJECT}",
        proxies=setup_proxies(),  # type: ignore[arg-type]
    )


def create_new_session(region_name: Optional[str] = None, profile: Optional[str] = None) -> Session:
    return Session(region_name=region_name, profile_name=profile)


def create_new_session_with_creds(
    aws_access_key_id: str, aws_secret_access_key: str, aws_session_token: str, region_name: Optional[str] = None
) -> Session:
    return boto3.Session(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token,
        region_name=region_name,
    )


@overload
def boto3_client(
    service_name: Literal["codebuild"],
    session: Optional[Union[Callable[[], Session], Session]] = ...,
    region_name: Optional[str] = ...,
    profile: Optional[str] = ...,
    aws_access_key_id: Optional[str] = ...,
    aws_secret_access_key: Optional[str] = ...,
    aws_session_token: Optional[str] = ...,
) -> "CodeBuildClient": ...


@overload
def boto3_client(
    service_name: Literal["iam"],
    session: Optional[Union[Callable[[], Session], Session]] = ...,
    region_name: Optional[str] = ...,
    profile: Optional[str] = ...,
    aws_access_key_id: Optional[str] = ...,
    aws_secret_access_key: Optional[str] = ...,
    aws_session_token: Optional[str] = ...,
) -> "IAMClient": ...


@overload
def boto3_client(
    service_name: Literal["s3"],
    session: Optional[Union[Callable[[], Session], Session]] = ...,
    region_name: Optional[str] = ...,
    profile: Optional[str] = ...,
    aws_access_key_id: Optional[str] = ...,
    aws_secret_access_key: Optional[str] = ...,
    aws_session_token: Optional[str] = ...,
) -> "S3Client": ...


@overload
def boto3_client(
    service_name: Literal["secretsmanager"],
    session: Optional[Union[Callable[[], Session], Session]] = ...,
    region_name: Optional[str] = ...,
    profile: Optional[str] = ...,
    aws_access_key_id: Optional[str] = ...,
    aws_secret_access_key: Optional[str] = ...,
    aws_session_token: Optional[str] = ...,
) -> "SecretsManagerClient": ...


@overload
def boto3_client(
    service_name: Literal["ssm"],
    session: Optional[Union[Callable[[], Session], Session]] = ...,
    region_name: Optional[str] = ...,
    profile: Optional[str] = ...,
    aws_access_key_id: Optional[str] = ...,
    aws_secret_access_key: Optional[str] = ...,
    aws_session_token: Optional[str] = ...,
) -> "SSMClient": ...


@overload
def boto3_client(
    service_name: Literal["sts"],
    session: Optional[Union[Callable[[], Session], Session]] = ...,
    region_name: Optional[str] = ...,
    profile: Optional[str] = ...,
    aws_access_key_id: Optional[str] = ...,
    aws_secret_access_key: Optional[str] = ...,
    aws_session_token: Optional[str] = ...,
) -> "STSClient": ...


@overload
def boto3_client(
    service_name: Literal["cloudformation"],
    session: Optional[Union[Callable[[], Session], Session]] = ...,
    region_name: Optional[str] = ...,
    profile: Optional[str] = ...,
    aws_access_key_id: Optional[str] = ...,
    aws_secret_access_key: Optional[str] = ...,
    aws_session_token: Optional[str] = ...,
) -> "CloudFormationClient": ...


@overload
def boto3_client(
    service_name: Literal["logs"],
    session: Optional[Union[Callable[[], Session], Session]] = ...,
    region_name: Optional[str] = ...,
    profile: Optional[str] = ...,
    aws_access_key_id: Optional[str] = ...,
    aws_secret_access_key: Optional[str] = ...,
    aws_session_token: Optional[str] = ...,
) -> "CloudWatchLogsClient": ...


@overload
def boto3_client(
    service_name: str,
    session: Optional[Union[Callable[[], Session], Session]] = ...,
    region_name: Optional[str] = ...,
    profile: Optional[str] = ...,
    aws_access_key_id: Optional[str] = ...,
    aws_secret_access_key: Optional[str] = ...,
    aws_session_token: Optional[str] = ...,
) -> "BaseClient": ...


def boto3_client(
    service_name: str,
    session: Optional[Union[Callable[[], Session], Session]] = None,
    region_name: Optional[str] = None,
    profile: Optional[str] = None,
    aws_access_key_id: Optional[str] = None,
    aws_secret_access_key: Optional[str] = None,
    aws_session_token: Optional[str] = None,
) -> "BaseClient":
    if aws_access_key_id and aws_secret_access_key and aws_session_token:
        return create_new_session_with_creds(  # type: ignore[call-overload,no-any-return]
            aws_access_key_id, aws_secret_access_key, aws_session_token, region_name
        ).client(service_name=service_name, use_ssl=True, config=get_botocore_config())
    elif not session:
        return create_new_session(region_name, profile).client(  # type: ignore[call-overload,no-any-return]
            service_name=service_name, use_ssl=True, config=get_botocore_config()
        )
    else:
        if isinstance(session, Session):
            return session.client(service_name=service_name, use_ssl=True, config=get_botocore_config())  # type: ignore
        else:
            raise TypeError(f"Expected boto3.Session instance, got {type(session)}")


@overload
def boto3_resource(
    service_name: Literal["iam"],
    ssession: Optional[Union[Callable[[], Session], Session]] = ...,
    region_name: Optional[str] = ...,
    profile: Optional[str] = ...,
) -> "IAMServiceResource": ...


@overload
def boto3_resource(
    service_name: Literal["s3"],
    ssession: Optional[Union[Callable[[], Session], Session]] = ...,
    region_name: Optional[str] = ...,
    profile: Optional[str] = ...,
) -> "S3ServiceResource": ...


@overload
def boto3_resource(
    service_name: str,
    session: Optional[Union[Callable[[], Session], Session]] = ...,
    region_name: Optional[str] = ...,
    profile: Optional[str] = ...,
) -> "ServiceResource": ...


def boto3_resource(  # type: ignore[misc]
    service_name: str,
    session: Optional[Union[Callable[[], Session], Session]] = None,
    region_name: Optional[str] = None,
    profile: Optional[str] = None,
) -> "ServiceResource":
    if not session:
        return create_new_session(region_name=region_name, profile=profile).resource(  # type: ignore[no-any-return]
            service_name=service_name, use_ssl=True, config=get_botocore_config()
        )  # type: ignore[call-overload]
    else:
        if callable(session):
            session = session()
        return session.resource(  # type: ignore[no-any-return]
            service_name=service_name,
            use_ssl=True,
            config=get_botocore_config(),
        )  # type: ignore[call-overload]


def get_region(session: Optional[Union[Callable[[], Session], Session]] = None, profile: Optional[str] = None) -> str:
    sess = session() if callable(session) else (session or create_new_session(profile=profile))
    if sess.region_name is None:
        raise seedfarmer.errors.InvalidConfigurationError(
            "It is not possible to infer AWS REGION from your environment."
        )
    return str(sess.region_name)


def _call_sts(
    session: Optional[Union[Callable[[], Session], Session]] = None, profile: Optional[str] = None
) -> Dict[str, Any]:
    try:
        if not session:
            return cast(Dict[str, Any], boto3_client(service_name="sts", profile=profile).get_caller_identity())
        else:
            return cast(Dict[str, Any], boto3_client(service_name="sts", session=session).get_caller_identity())
    except botocore.exceptions.NoCredentialsError as e:
        _logger.error(f"ERROR: {e}")
        from seedfarmer.output_utils import print_bolded

        print_bolded("Please make sure you have valid AWS Credentials", color="red")
        raise e

    except botocore.exceptions.ClientError as e:
        _logger.error(f"ERROR: {e}")
        from seedfarmer.output_utils import print_bolded

        print_bolded("Please make sure you have a valid AWS Session", color="red")
        raise e


def get_sts_identity_info(
    session: Optional[Union[Callable[[], Session], Session]] = None, profile: Optional[str] = None
) -> Tuple[str, str, str]:
    sts_info = _call_sts(session=session, profile=profile)
    return cast(
        Tuple[str, str, str], (sts_info.get("Account"), sts_info.get("Arn"), str(sts_info.get("Arn")).split(":")[1])
    )


def create_signed_request(
    endpoint: str,
    session: Session,
    credentials: Credentials,
    service: str = "s3",
    region: Optional[str] = None,
    method: Optional[str] = "GET",
    params: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, Any]] = None,
) -> AWSRequest:
    region = get_region(session) if not region else region
    auth = SigV4Auth(credentials, service, region)

    request = AWSRequest(method=method, url=endpoint, params=params, data=None, headers=headers)
    auth.add_auth(request)

    return request


def try_it(
    f: Callable[..., Any],
    ex: Any,
    base: float = 1.0,
    max_num_tries: int = 3,
    **kwargs: Any,
) -> Any:
    """Run function with decorrelated Jitter.

    Reference: https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    """
    delay: float = base
    for i in range(max_num_tries):
        try:
            return f(**kwargs)
        except ex as exception:
            if i == (max_num_tries - 1):
                raise exception
            delay = random.uniform(base, delay * 3)
            _logger.error(
                "Retrying %s | Fail number %s/%s | Exception: %s",
                f,
                i + 1,
                max_num_tries,
                exception,
            )
            time.sleep(delay)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import glob
import hashlib
import logging
import os
import re
import shutil
from typing import Any, Dict, List, Optional, Union

import humps
import yaml
from boto3 import Session
from dotenv import dotenv_values, load_dotenv
from ruamel.yaml import YAML
from ruamel.yaml.scalarstring import PreservedScalarString

import seedfarmer.errors
from seedfarmer.services._service_utils import get_region, get_sts_identity_info

_logger: logging.Logger = logging.getLogger(__name__)

NoDatesSafeLoader = yaml.SafeLoader


class CfnSafeYamlLoader(yaml.SafeLoader):
    """
    CfnSafeYamlLoader
    A predefined class loader to support safely reading YAML files

    Parameters
    ----------
    yaml : _type_
    """

    yaml_implicit_resolvers = {
        k: [r for r in v if r[0] != "tag:yaml.org,2002:timestamp"]
        for k, v in NoDatesSafeLoader.yaml_implicit_resolvers.items()
    }


class LiteralStr(PreservedScalarString):
    """A string subclass that forces block style (|) YAML formatting."""

    pass


def register_literal_str() -> YAML:
    yaml = YAML()
    yaml.preserve_quotes = True
    yaml.indent(mapping=2, sequence=4, offset=2)
    yaml.block_seq_indent = 2

    def literal_str_representer(dumper, data):  # type: ignore [no-untyped-def]
        kwargs = {"style": "|"}
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, **kwargs)

    yaml.representer.add_representer(LiteralStr, literal_str_representer)
    return yaml


def apply_literalstr(s: str) -> Union[str, LiteralStr]:
    return LiteralStr(s) if "\n" in s else s


def upper_snake_case(value: str) -> str:
    """
    This will convert strings to a standard format

    Parameters
    ----------
    value : str
        The string you want to convert

    Returns
    -------
    str
        the string standardized
    """
    if humps.is_camelcase(value):
        return humps.decamelize(value).upper()
    elif humps.is_pascalcase(value):
        return humps.depascalize(value).upper()
    else:
        return value.replace("-", "_").upper()


def generate_hash(string: str, length: int = 8) -> str:
    return (hashlib.sha1(string.encode("UTF-8"), usedforsecurity=False).hexdigest())[:length]


def generate_session_hash(session: Optional[Session] = None) -> str:
    """
    Generate a hexdigest hash of the project and the deployment - for use generating unique names

    Returns
    -------
    str
        The resulting hash as a string
    """
    account, _, _ = get_sts_identity_info(session=session)
    region = get_region(session=session)
    concatenated_string = f"{account}-{region}"
    hash_value = generate_hash(string=concatenated_string, length=8)
    _logger.debug("HASH generated is %s", hash_value)
    return hash_value


def generate_codebuild_url(account_id: str, region: str, codebuild_id: str, partition: Optional[str] = "aws") -> str:
    """
    Generate a standard URL for codebuild build information

    Parameters
    ----------
    account_id : str
        The AWS account id where CodeBuild ran
    region : str
        The AWS region where CodeBuild ran
    codebuild_id : str
        The CodeBuild Build ID

    Returns
    -------
    str
        The standard URL with protocol and query parameters
        ex: https://us-east-1.console.aws.amazon.com/codesuite/codebuild/123456789012/projects/
            codebuild-id/build/codebuild-id:3413241234/?region-us-east-1 (LEGACY)
        if in a differing partition (ex.aws-cn) the url looks like:
            https://cn-north-1.console.amazonaws.cn/codesuite/codebuild/123456789012/projects/
            codeseeder-idf/build/codeseeder-id:3413241234/?region=cn-north-1 (LEGACY)
    """
    try:
        b_id_enc = codebuild_id.replace(":", "%3A")
        cb_p = codebuild_id.split(":")[0]
        domain_completion = ".console.aws.amazon.com/codesuite/codebuild/"
        if partition == "aws-cn":
            domain_completion = ".console.amazonaws.cn/codesuite/codebuild/"
        return "".join(
            (
                "https://",
                f"{region}",
                f"{domain_completion}",
                f"{account_id}",
                "/projects/",
                f"{cb_p}",
                "/build/",
                f"{b_id_enc}",
                "/?region=",
                f"{region}",
            )
        )
    except Exception as e:
        _logger.error(f"Error...{account_id} - {region} - {codebuild_id} - {e} ")
        return "N/A"


def get_toolchain_role_name(project_name: str, qualifier: Optional[str] = None) -> str:
    name = f"seedfarmer-{project_name}-toolchain-role"
    return f"{name}-{qualifier}" if qualifier else name


def get_toolchain_role_arn(
    partition: str,
    toolchain_account_id: str,
    project_name: str,
    qualifier: Optional[str] = None,
    role_prefix: Optional[str] = None,
) -> str:
    role_prefix = role_prefix if role_prefix else "/"
    return (
        f"arn:{partition}:iam::{toolchain_account_id}:role{role_prefix}"
        f"{get_toolchain_role_name(project_name, qualifier)}"
    )


def get_deployment_role_name(project_name: str, qualifier: Optional[str] = None) -> str:
    name = f"seedfarmer-{project_name}-deployment-role"
    return f"{name}-{qualifier}" if qualifier else name


def get_deployment_role_arn(
    partition: str,
    deployment_account_id: str,
    project_name: str,
    qualifier: Optional[str] = None,
    role_prefix: Optional[str] = None,
) -> str:
    role_prefix = role_prefix if role_prefix else "/"
    return (
        f"arn:{partition}:iam::{deployment_account_id}:role{role_prefix}"
        f"{get_deployment_role_name(project_name, qualifier)}"
    )


def get_generic_module_deployment_role_name(
    project_name: str,
    deployment_name: str,
    region: str,
) -> str:
    resource_name = f"{project_name}-{deployment_name}-{region}"
    resource_hash = generate_hash(string=resource_name, length=4)

    # Max length of IAM role name is 64 chars, "-deployment-role" is 16 chars, resource_hash plus "-" is 5 chars.
    # If the resource_name, and "-deployment-role" is too long, truncate and use resource_hash for uniqueness.
    return (
        f"{resource_name[: 64 - 16 - 5]}-deployment-role-{resource_hash}"
        if len(resource_name) > (64 - 16)
        else f"{resource_name}-deployment-role"
    )


def valid_qualifier(qualifer: str) -> bool:
    return True if ((len(qualifer) <= 6) and qualifer.isalnum()) else False


def load_dotenv_files(root_path: str, env_files: List[str]) -> None:
    """
    Load the environment variables from the .env files

    Parameters
    ----------
    root_path : str
        The path to the root of the project
    env_files : List[str]
        The list of the .env files to load
    """
    loaded_values = {}

    for env_file in env_files:
        _logger.info("Loading environment variables from %s", env_file)
        try:
            dotenv_path = os.path.join(root_path, env_file)
            _logger.debug("Loading environment variables from %s", dotenv_path)
            load_dotenv(dotenv_path=dotenv_path, verbose=True, override=True)
        except FileNotFoundError:
            _logger.error("Error loading .env file")
        try:
            loaded_values.update(dotenv_values(dotenv_path, verbose=True))
        except Exception as e:
            _logger.error(f"Failed to load environment variables from {env_file}: {e}")

    _logger.debug("Loaded environment variables: %s", loaded_values)


def remove_nulls(payload: Dict[str, Any]) -> Dict[str, Any]:
    if isinstance(payload, dict):
        return {k: remove_nulls(v) for k, v in payload.items() if v is not None}
    elif isinstance(payload, list):
        return [remove_nulls(v) for v in payload]
    else:
        return payload


def batch_replace_env(payload: Dict[str, Any]) -> Dict[str, Any]:
    pattern = r"\${(.*?)}"

    def replace_str(value: str) -> str:
        try:
            return re.sub(
                pattern=pattern,
                repl=lambda m: os.environ[m.group(1).strip()],
                string=value,
            )
        except KeyError as e:
            # Show just the variable name, not the full KeyError
            var_name = str(e).strip("'\"")
            raise seedfarmer.errors.InvalidManifestError(f"Environment variable not set: {var_name}")

    def recurse_list(working_list: List[Any]) -> List[Any]:
        for key, value in enumerate(working_list):
            if isinstance(value, str):
                working_list[key] = replace_str(value)
            elif isinstance(value, list):
                recurse_list(value)
            elif isinstance(value, dict):
                recurse_dict(value)
        return working_list

    def recurse_dict(working_element: Dict[str, Any]) -> Dict[str, Any]:
        if not working_element.get("disableEnvVarResolution"):
            for key, value in working_element.items():
                if isinstance(value, str):
                    working_element[key] = replace_str(value)
                elif isinstance(value, list):
                    recurse_list(value)
                elif isinstance(value, dict) and key not in ["deploy_spec"]:
                    recurse_dict(value)
        return working_element

    payload = recurse_dict(payload)
    return payload


def create_output_dir(name: str, path_override: Optional[str] = None) -> str:
    """Helper function for creating or clearing a .seedfarmer.out output directory by default

    Parameters
    ----------
    name : str
        Name of the directory to create in  the .seedfarmer.out directory

    path_override: Optional[str]
        If you want to override the name .seedfarmer.out use this (beware of what you are doing)

    Returns
    -------
    str
        Full path of the created directory
    """
    local_path = path_override if path_override else ".seedfarmer.out"
    out_dir = os.path.join(os.getcwd(), local_path, name)
    try:
        shutil.rmtree(out_dir)
    except FileNotFoundError:
        pass
    # except PermissionError:
    #     pass
    os.makedirs(out_dir, exist_ok=True)
    return out_dir


def delete_all_output_dir(name: str = ".seedfarmerlocal-") -> None:
    pattern = os.path.join(os.getcwd(), f"{name}-*")
    for path in glob.glob(pattern):
        if os.path.isdir(path):
            shutil.rmtree(path)
